from mpmath import mp, workprec

from ...core import (Add, Dummy, EulerGamma, Expr, Function, I, Integer, Pow,
                     Rational, oo, pi, sympify, zoo)
from ...core.function import ArgumentIndexError
from ..combinatorial.factorials import RisingFactorial, factorial, rf
from ..combinatorial.numbers import bernoulli, harmonic
from ..elementary.exponential import exp, log
from ..elementary.integers import ceiling, floor
from ..elementary.miscellaneous import sqrt
from .error_functions import erf
from .zeta_functions import zeta


###############################################################################
# ########################## COMPLETE GAMMA FUNCTION ######################## #
###############################################################################

class gamma(Function):
    r"""
    The gamma function

    .. math::
        \Gamma(x) := \int^{\infty}_{0} t^{x-1} e^{t} \mathrm{d}t.

    The ``gamma`` function implements the function which passes through the
    values of the factorial function, i.e. `\Gamma(n) = (n - 1)!` when n is
    an integer. More general, `\Gamma(z)` is defined in the whole complex
    plane except at the negative integers where there are simple poles.

    Examples
    ========

    Several special values are known:

    >>> gamma(1)
    1
    >>> gamma(4)
    6
    >>> gamma(Rational(3, 2))
    sqrt(pi)/2

    The Gamma function obeys the mirror symmetry:

    >>> conjugate(gamma(x))
    gamma(conjugate(x))

    Differentiation with respect to x is supported:

    >>> diff(gamma(x), x)
    gamma(x)*polygamma(0, x)

    Series expansion is also supported:

    >>> series(gamma(x), x, 0, 3)
    1/x - EulerGamma + x*(EulerGamma**2/2 + pi**2/12) + x**2*(-EulerGamma*pi**2/12 + polygamma(2, 1)/6 - EulerGamma**3/6) + O(x**3)

    We can numerically evaluate the gamma function to arbitrary precision
    on the whole complex plane:

    >>> gamma(pi).evalf(40)
    2.288037795340032417959588909060233922890
    >>> gamma(1+I).evalf(20)
    0.49801566811835604271 - 0.15494982830181068512*I

    See Also
    ========

    lowergamma: Lower incomplete gamma function.
    uppergamma: Upper incomplete gamma function.
    polygamma: Polygamma function.
    loggamma: Log Gamma function.
    digamma: Digamma function.
    trigamma: Trigamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Gamma_function
    * https://dlmf.nist.gov/5
    * https://mathworld.wolfram.com/GammaFunction.html
    * http://functions.wolfram.com/GammaBetaErf/Gamma/

    """

    unbranched = True

    def fdiff(self, argindex=1):
        if argindex == 1:
            return self.func(self.args[0])*polygamma(0, self.args[0])
        else:
            raise ArgumentIndexError(self, argindex)

    @classmethod
    def eval(cls, arg):
        if arg.is_Number:
            if arg is oo:
                return oo
            elif arg.is_Integer:
                if arg.is_positive:
                    return factorial(arg - 1)
                else:
                    return zoo
            elif arg.is_Rational:
                if arg.denominator == 2:
                    n = abs(arg.numerator) // arg.denominator

                    if arg.is_positive:
                        k, coeff = n, Integer(1)
                    else:
                        n = k = n + 1

                        if n & 1 == 0:
                            coeff = Integer(1)
                        else:
                            coeff = Integer(-1)

                    for i in range(3, 2*k, 2):
                        coeff *= i

                    if arg.is_positive:
                        return coeff*sqrt(pi) / 2**n
                    else:
                        return 2**n*sqrt(pi) / coeff

        if arg.is_integer and arg.is_nonpositive:
            return zoo

    def _eval_expand_func(self, **hints):
        arg = self.args[0]
        if arg.is_Rational:
            if abs(arg.numerator) > arg.denominator:
                x = Dummy('x')
                n = arg.numerator // arg.denominator
                p = arg.numerator - n*arg.denominator
                return self.func(x + n)._eval_expand_func().subs({x: Rational(p, arg.denominator)})

        if arg.is_Add:
            coeff, tail = arg.as_coeff_add()
            if coeff and coeff.denominator != 1:
                intpart = floor(coeff)
                tail = (coeff - intpart,) + tail
                coeff = intpart
            tail = arg._new_rawargs(*tail, reeval=False)
            return self.func(tail)*RisingFactorial(tail, coeff)

        return self.func(*self.args)

    def _eval_conjugate(self):
        return self.func(self.args[0].conjugate())

    def _eval_is_extended_real(self):
        x = self.args[0]
        if x.is_positive or x.is_noninteger:
            return True

    def _eval_is_positive(self):
        x = self.args[0]
        if x.is_positive:
            return True
        elif x.is_noninteger:
            return floor(x).is_even

    def _eval_rewrite_as_tractable(self, z):
        return exp(loggamma(z))

    def _eval_rewrite_as_factorial(self, z):
        return factorial(z - 1)

    def _eval_nseries(self, x, n, logx):
        x0 = self.args[0].limit(x, 0)
        if not (x0.is_Integer and x0 <= 0):
            return super()._eval_nseries(x, n, logx)
        t = self.args[0] - x0
        return (self.func(t + 1)/rf(self.args[0], -x0 + 1))._eval_nseries(x, n, logx)

    def _latex(self, printer, exp=None):
        aa = printer._print(self.args[0])
        if exp:
            return r'\Gamma^{%s}{\left(%s \right)}' % (printer._print(exp), aa)
        else:
            return r'\Gamma{\left(%s \right)}' % aa

    @staticmethod
    def _latex_no_arg(printer):
        return r'\Gamma'


###############################################################################
# ################ LOWER and UPPER INCOMPLETE GAMMA FUNCTIONS ############### #
###############################################################################

class lowergamma(Function):
    r"""
    The lower incomplete gamma function.

    It can be defined as the meromorphic continuation of

    .. math::
        \gamma(s, x) := \int_0^x t^{s-1} e^{-t} \mathrm{d}t = \Gamma(s) - \Gamma(s, x).

    This can be shown to be the same as

    .. math::
        \gamma(s, x) = \frac{x^s}{s} {}_1F_1\left({s \atop s+1} \middle| -x\right),

    where `{}_1F_1` is the (confluent) hypergeometric function.

    Examples
    ========

    >>> from diofant.abc import s
    >>> lowergamma(s, x)
    lowergamma(s, x)
    >>> lowergamma(3, x)
    2 - E**(-x)*x**2 - 2*E**(-x)*x - 2*E**(-x)
    >>> lowergamma(-Rational(1, 2), x)
    -2*sqrt(pi)*erf(sqrt(x)) - 2*E**(-x)/sqrt(x)

    See Also
    ========

    gamma: Gamma function.
    uppergamma: Upper incomplete gamma function.
    polygamma: Polygamma function.
    loggamma: Log Gamma function.
    digamma: Digamma function.
    trigamma: Trigamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Incomplete_gamma_function#Lower_incomplete_Gamma_function
    * Abramowitz, Milton; Stegun, Irene A., eds. (1965), Chapter 6, Section 5,
      Handbook of Mathematical Functions with Formulas, Graphs, and Mathematical Tables
    * https://dlmf.nist.gov/8
    * http://functions.wolfram.com/GammaBetaErf/Gamma2/
    * http://functions.wolfram.com/GammaBetaErf/Gamma3/

    """

    def fdiff(self, argindex=2):
        from .hyper import meijerg
        from .. import unpolarify
        if argindex == 2:
            a, z = self.args
            return exp(-unpolarify(z))*z**(a - 1)
        elif argindex == 1:
            a, z = self.args
            return gamma(a)*digamma(a) - log(z)*uppergamma(a, z) \
                - meijerg([], [1, 1], [0, 0, a], [], z)

        else:
            raise ArgumentIndexError(self, argindex)

    @classmethod
    def eval(cls, a, x):
        # For lack of a better place, we use this one to extract branching
        # information. The following can be
        # found in the literature (c/f references given above), albeit scattered:
        # 1) For fixed x != 0, lowergamma(s, x) is an entire function of s
        # 2) For fixed positive integers s, lowergamma(s, x) is an entire
        #    function of x.
        # 3) For fixed non-positive integers s,
        #    lowergamma(s, exp(I*2*pi*n)*x) =
        #              2*pi*I*n*(-1)**(-s)/factorial(-s) + lowergamma(s, x)
        #    (this follows from lowergamma(s, x).diff(x) = x**(s-1)*exp(-x)).
        # 4) For fixed non-integral s,
        #    lowergamma(s, x) = x**s*gamma(s)*lowergamma_unbranched(s, x),
        #    where lowergamma_unbranched(s, x) is an entire function (in fact
        #    of both s and x), i.e.
        #    lowergamma(s, exp(2*I*pi*n)*x) = exp(2*pi*I*n*a)*lowergamma(a, x)
        from .. import unpolarify
        nx, n = x.extract_branch_factor()
        if a.is_integer and a.is_positive:
            nx = unpolarify(x)
            if nx != x:
                return lowergamma(a, nx)
        elif a.is_integer and a.is_nonpositive:
            if n != 0:
                return 2*pi*I*n*(-1)**(-a)/factorial(-a) + lowergamma(a, nx)
        elif n != 0:
            return exp(2*pi*I*n*a)*lowergamma(a, nx)

        # Special values.
        if a.is_Number:
            # TODO this should be non-recursive
            if a == 1:
                return 1 - exp(-x)
            elif a == Rational(1, 2):
                return sqrt(pi)*erf(sqrt(x))
            elif a.is_Integer or (2*a).is_Integer:
                b = a - 1
                if b.is_positive:
                    return b*cls(b, x) - x**b * exp(-x)

                if not a.is_Integer:
                    return (cls(a + 1, x) + x**a * exp(-x))/a

    def _eval_evalf(self, prec):
        a = self.args[0]._to_mpmath(prec)
        z = self.args[1]._to_mpmath(prec)
        with workprec(prec):
            res = mp.gammainc(a, 0, z)
        return Expr._from_mpmath(res, prec)

    def _eval_conjugate(self):
        z = self.args[1]
        if z not in (0, -oo):
            return self.func(self.args[0].conjugate(), z.conjugate())

    def _eval_rewrite_as_uppergamma(self, s, x):
        return gamma(s) - uppergamma(s, x)

    def _eval_rewrite_as_expint(self, s, x):
        from .error_functions import expint
        if s.is_integer and s.is_nonpositive:
            return self
        return self.rewrite(uppergamma).rewrite(expint)

    @staticmethod
    def _latex_no_arg(printer):
        return r'\gamma'


class uppergamma(Function):
    r"""
    The upper incomplete gamma function.

    It can be defined as the meromorphic continuation of

    .. math::
        \Gamma(s, x) := \int_x^\infty t^{s-1} e^{-t} \mathrm{d}t = \Gamma(s) - \gamma(s, x).

    where `\gamma(s, x)` is the lower incomplete gamma function,
    :class:`lowergamma`. This can be shown to be the same as

    .. math::
        \Gamma(s, x) = \Gamma(s) - \frac{x^s}{s} {}_1F_1\left({s \atop s+1} \middle| -x\right),

    where `{}_1F_1` is the (confluent) hypergeometric function.

    The upper incomplete gamma function is also essentially equivalent to the
    generalized exponential integral:

    .. math::
        \operatorname{E}_{n}(x) = \int_{1}^{\infty}{\frac{e^{-xt}}{t^n} \, dt} = x^{n-1}\Gamma(1-n,x).

    Examples
    ========

    >>> from diofant.abc import s
    >>> uppergamma(s, x)
    uppergamma(s, x)
    >>> uppergamma(3, x)
    E**(-x)*x**2 + 2*E**(-x)*x + 2*E**(-x)
    >>> uppergamma(-Rational(1, 2), x)
    -2*sqrt(pi)*(-erf(sqrt(x)) + 1) + 2*E**(-x)/sqrt(x)
    >>> uppergamma(-2, x)
    expint(3, x)/x**2

    See Also
    ========

    gamma: Gamma function.
    lowergamma: Lower incomplete gamma function.
    polygamma: Polygamma function.
    loggamma: Log Gamma function.
    digamma: Digamma function.
    trigamma: Trigamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Incomplete_gamma_function#Upper_incomplete_Gamma_function
    * Abramowitz, Milton; Stegun, Irene A., eds. (1965), Chapter 6, Section 5,
      Handbook of Mathematical Functions with Formulas, Graphs, and Mathematical Tables
    * https://dlmf.nist.gov/8
    * http://functions.wolfram.com/GammaBetaErf/Gamma2/
    * http://functions.wolfram.com/GammaBetaErf/Gamma3/
    * https://en.wikipedia.org/wiki/Exponential_integral#Relation_with_other_functions

    """

    def fdiff(self, argindex=2):
        from .hyper import meijerg
        from .. import unpolarify
        if argindex == 2:
            a, z = self.args
            return -exp(-unpolarify(z))*z**(a - 1)
        elif argindex == 1:
            a, z = self.args
            return uppergamma(a, z)*log(z) + meijerg([], [1, 1], [0, 0, a], [], z)
        else:
            raise ArgumentIndexError(self, argindex)

    def _eval_evalf(self, prec):
        a = self.args[0]._to_mpmath(prec)
        z = self.args[1]._to_mpmath(prec)
        with workprec(prec):
            res = mp.gammainc(a, z, mp.inf)
        return Expr._from_mpmath(res, prec)

    @classmethod
    def eval(cls, a, z):
        from .error_functions import expint
        from .. import unpolarify
        if z.is_Number:
            if z is oo:
                return Integer(0)
            elif z == 0:
                # TODO: Holds only for Re(a) > 0:
                return gamma(a)

        # We extract branching information here. C/f lowergamma.
        nx, n = z.extract_branch_factor()
        if a.is_integer and a.is_positive:
            nx = unpolarify(z)
            if z != nx:
                return uppergamma(a, nx)
        elif a.is_integer and a.is_nonpositive:
            if n != 0:
                return -2*pi*I*n*(-1)**(-a)/factorial(-a) + uppergamma(a, nx)
        elif n != 0:
            return gamma(a)*(1 - exp(2*pi*I*n*a)) + exp(2*pi*I*n*a)*uppergamma(a, nx)

        # Special values.
        if a.is_Number:
            # TODO this should be non-recursive
            if a == 1:
                return exp(-z)
            elif a == Rational(1, 2):
                return sqrt(pi)*(1 - erf(sqrt(z)))  # TODO could use erfc...
            elif a.is_Integer or (2*a).is_Integer:
                b = a - 1
                if b.is_positive:
                    return b*cls(b, z) + z**b * exp(-z)
                elif b.is_Integer:
                    return expint(-b, z)*unpolarify(z)**(b + 1)
                else:
                    return (cls(a + 1, z) - z**a * exp(-z))/a

    def _eval_conjugate(self):
        z = self.args[1]
        if z not in (0, -oo):
            return self.func(self.args[0].conjugate(), z.conjugate())

    def _eval_rewrite_as_lowergamma(self, s, x):
        return gamma(s) - lowergamma(s, x)

    def _eval_rewrite_as_expint(self, s, x):
        from .error_functions import expint
        return expint(1 - s, x)*x**s


###############################################################################
# #################### POLYGAMMA and LOGGAMMA FUNCTIONS ##################### #
###############################################################################

class polygamma(Function):
    r"""
    The function ``polygamma(n, z)`` returns ``log(gamma(z)).diff(n + 1)``.

    It is a meromorphic function on `\mathbb{C}` and defined as the (n+1)-th
    derivative of the logarithm of the gamma function:

    .. math::
        \psi^{(n)} (z) := \frac{\mathrm{d}^{n+1}}{\mathrm{d} z^{n+1}} \log\Gamma(z).

    Examples
    ========

    Several special values are known:

    >>> polygamma(0, 1)
    -EulerGamma
    >>> polygamma(0, Rational(1, 2))
    -2*log(2) - EulerGamma
    >>> polygamma(0, Rational(1, 3))
    -3*log(3)/2 - sqrt(3)*pi/6 - EulerGamma
    >>> polygamma(0, Rational(1, 4))
    -3*log(2) - pi/2 - EulerGamma
    >>> polygamma(0, 2)
    -EulerGamma + 1
    >>> polygamma(0, 23)
    -EulerGamma + 19093197/5173168

    >>> polygamma(0, oo)
    oo
    >>> polygamma(0, -oo)
    oo
    >>> polygamma(0, I*oo)
    oo
    >>> polygamma(0, -I*oo)
    oo

    Differentiation with respect to x is supported:

    >>> diff(polygamma(0, x), x)
    polygamma(1, x)
    >>> diff(polygamma(0, x), x, 2)
    polygamma(2, x)
    >>> diff(polygamma(0, x), x, 3)
    polygamma(3, x)
    >>> diff(polygamma(1, x), x)
    polygamma(2, x)
    >>> diff(polygamma(1, x), x, 2)
    polygamma(3, x)
    >>> diff(polygamma(2, x), x)
    polygamma(3, x)
    >>> diff(polygamma(2, x), x, 2)
    polygamma(4, x)

    >>> diff(polygamma(n, x), x)
    polygamma(n + 1, x)
    >>> diff(polygamma(n, x), x, 2)
    polygamma(n + 2, x)

    We can rewrite polygamma functions in terms of harmonic numbers:

    >>> polygamma(0, x).rewrite(harmonic)
    harmonic(x - 1) - EulerGamma
    >>> polygamma(2, x).rewrite(harmonic)
    2*harmonic(x - 1, 3) - 2*zeta(3)
    >>> ni = Symbol('n', integer=True)
    >>> polygamma(ni, x).rewrite(harmonic)
    (-1)**(n + 1)*(-harmonic(x - 1, n + 1) + zeta(n + 1))*factorial(n)

    See Also
    ========

    gamma: Gamma function.
    lowergamma: Lower incomplete gamma function.
    uppergamma: Upper incomplete gamma function.
    loggamma: Log Gamma function.
    digamma: Digamma function.
    trigamma: Trigamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Polygamma_function
    * https://mathworld.wolfram.com/PolygammaFunction.html
    * http://functions.wolfram.com/GammaBetaErf/PolyGamma/
    * http://functions.wolfram.com/GammaBetaErf/PolyGamma2/

    """

    def fdiff(self, argindex=2):
        if argindex == 2:
            n, z = self.args[:2]
            return polygamma(n + 1, z)
        else:
            raise ArgumentIndexError(self, argindex)

    def _eval_is_positive(self):
        if self.args[0].is_positive and self.args[1].is_positive:
            if self.args[0].is_odd:
                return True

    def _eval_aseries(self, n, args0, x, logx):
        from ...series import Order
        if args0[1] != oo or not \
                (self.args[0].is_Integer and self.args[0].is_nonnegative):
            return super()._eval_aseries(n, args0, x, logx)
        z = self.args[1]
        N = self.args[0]

        if N == 0:
            # digamma function series
            # Abramowitz & Stegun, p. 259, 6.3.18
            r = log(z) - 1/(2*z)
            o = None
            if n < 2:
                o = Order(1/z, x)
            else:
                m = ceiling((n + 1)//2)
                l = [bernoulli(2*k) / (2*k*z**(2*k)) for k in range(1, m)]
                r -= Add(*l)
                o = Order(1/z**(2*m), x)
            return r._eval_nseries(x, n, logx) + o
        else:
            # proper polygamma function
            # Abramowitz & Stegun, p. 260, 6.4.10
            # We return terms to order higher than O(x**n) on purpose
            # -- otherwise we would not be able to return any terms for
            #    quite a long time!
            fac = gamma(N)
            e0 = fac + N*fac/(2*z)
            m = ceiling((n + 1)//2)
            for k in range(1, m):
                fac = fac*(2*k + N - 1)*(2*k + N - 2) / ((2*k)*(2*k - 1))
                e0 += bernoulli(2*k)*fac/z**(2*k)
            o = Order(1/z**(2*m), x)
            if n == 0:
                o = Order(1/z, x)
            elif n == 1:
                o = Order(1/z**2, x)
            r = e0 + o
            return (-1 * (-1/z)**N * r)._eval_nseries(x, n, logx)

    @classmethod
    def eval(cls, n, z):
        n, z = list(map(sympify, (n, z)))
        from .. import unpolarify

        if n.is_integer:
            if n.is_nonnegative:
                nz = unpolarify(z)
                if z != nz:
                    return polygamma(n, nz)

            if n == -1:
                return loggamma(z)
            else:
                if z.is_Number:
                    if z is oo:
                        if n.is_Number:
                            if n == 0:
                                return oo
                            else:
                                return Integer(0)
                    elif z.is_Integer:
                        if z.is_nonpositive:
                            return zoo
                        else:
                            if n == 0:
                                return -EulerGamma + harmonic(z - 1, 1)
                            elif n.is_odd:
                                return (-1)**(n + 1)*factorial(n)*zeta(n + 1, z)

        if n == 0:
            if z.is_Rational:
                # TODO actually *any* n/m can be done, but that is messy
                lookup = {Rational(1, 2): -2*log(2) - EulerGamma,
                          Rational(1, 3): -pi/2/sqrt(3) - 3*log(3)/2 - EulerGamma,
                          Rational(1, 4): -pi/2 - 3*log(2) - EulerGamma,
                          Rational(3, 4): -3*log(2) - EulerGamma + pi/2,
                          Rational(2, 3): -3*log(3)/2 + pi/2/sqrt(3) - EulerGamma}
                if z > 0:
                    n = floor(z)
                    z0 = z - n
                    if z0 in lookup:
                        return lookup[z0] + Add(*[1/(z0 + k) for k in range(n)])
                else:  # z < 0
                    n = floor(1 - z)
                    z0 = z + n
                    if z0 in lookup:
                        return lookup[z0] - Add(*[1/(z0 - 1 - k) for k in range(n)])
            elif z in (oo, -oo):
                return oo
            else:
                t = z.extract_multiplicatively(I)
                if t in (oo, -oo):
                    return oo

        # TODO n == 1 also can do some rational z

    def _eval_expand_func(self, **hints):
        n, z = self.args

        if n.is_Integer and n.is_nonnegative:
            if z.is_Add:
                coeff = z.args[0]
                if coeff.is_Integer:
                    e = -(n + 1)
                    if coeff > 0:
                        tail = Add(*[Pow(z - i, e)
                                     for i in range(1, int(coeff) + 1)])
                    else:
                        tail = -Add(*[Pow(z + i, e)
                                      for i in range(int(-coeff))])
                    return polygamma(n, z - coeff) + (-1)**n*factorial(n)*tail

            elif z.is_Mul:
                coeff, z = z.as_two_terms()
                if coeff.is_Integer and coeff.is_positive:
                    tail = [polygamma(n, z + Rational(i, coeff))
                            for i in range(int(coeff))]
                    if n == 0:
                        return Add(*tail)/coeff + log(coeff)
                    else:
                        return Add(*tail)/coeff**(n + 1)
                z *= coeff

        return polygamma(n, z)

    def _eval_rewrite_as_zeta(self, n, z):
        if (n - 1).is_nonnegative:
            return (-1)**(n + 1)*factorial(n)*zeta(n + 1, z)
        else:
            return self

    def _eval_rewrite_as_harmonic(self, n, z):
        if n.is_integer:
            if n == 0:
                return harmonic(z - 1) - EulerGamma
            else:
                return (-1)**(n+1) * factorial(n) * (zeta(n+1) - harmonic(z-1, n+1))

    def _eval_as_leading_term(self, x):
        from ...series import Order
        n, z = [a.as_leading_term(x) for a in self.args]
        o = Order(z, x)
        if n == 0 and o.contains(1/x):
            return o.getn() * log(x)
        else:
            return self.func(n, z)


class loggamma(Function):
    r"""
    The ``loggamma`` function implements the logarithm of the
    gamma function i.e, `\log\Gamma(x)`.

    Examples
    ========

    Several special values are known. For numerical integral
    arguments we have:

    >>> loggamma(-2)
    oo
    >>> loggamma(0)
    oo
    >>> loggamma(1)
    0
    >>> loggamma(2)
    0
    >>> loggamma(3)
    log(2)

    and for symbolic values:

    >>> n = Symbol('n', integer=True, positive=True)
    >>> loggamma(n)
    log(gamma(n))
    >>> loggamma(-n)
    oo

    for half-integral values:

    >>> loggamma(Rational(5, 2))
    log(3*sqrt(pi)/4)
    >>> loggamma(n/2)
    log(2**(-n + 1)*sqrt(pi)*gamma(n)/gamma(n/2 + 1/2))

    and general rational arguments:

    >>> L = loggamma(Rational(16, 3))
    >>> expand_func(L).doit()
    -5*log(3) + loggamma(1/3) + log(4) + log(7) + log(10) + log(13)
    >>> L = loggamma(Rational(19, 4))
    >>> expand_func(L).doit()
    -4*log(4) + loggamma(3/4) + log(3) + log(7) + log(11) + log(15)
    >>> L = loggamma(Rational(23, 7))
    >>> expand_func(L).doit()
    -3*log(7) + log(2) + loggamma(2/7) + log(9) + log(16)

    The loggamma function has the following limits towards infinity:

    >>> loggamma(oo)
    oo
    >>> loggamma(-oo)
    zoo

    The loggamma function obeys the mirror symmetry
    if `x \in \mathbb{C} \setminus \{-\infty, 0\}`:

    >>> c = Symbol('c', complex=True, real=False)
    >>> conjugate(loggamma(c))
    loggamma(conjugate(c))

    Differentiation with respect to x is supported:

    >>> diff(loggamma(x), x)
    polygamma(0, x)

    Series expansion is also supported:

    >>> series(loggamma(x), x, 0, 4)
    -log(x) - EulerGamma*x + pi**2*x**2/12 + x**3*polygamma(2, 1)/6 + O(x**4)

    We can numerically evaluate the gamma function to arbitrary precision
    on the whole complex plane:

    >>> loggamma(5).evalf(30)
    3.17805383034794561964694160130
    >>> loggamma(I).evalf(20)
    -0.65092319930185633889 - 1.8724366472624298171*I

    See Also
    ========

    gamma: Gamma function.
    lowergamma: Lower incomplete gamma function.
    uppergamma: Upper incomplete gamma function.
    polygamma: Polygamma function.
    digamma: Digamma function.
    trigamma: Trigamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Gamma_function
    * https://dlmf.nist.gov/5
    * https://mathworld.wolfram.com/LogGammaFunction.html
    * http://functions.wolfram.com/GammaBetaErf/LogGamma/

    """

    @classmethod
    def eval(cls, z):
        z = sympify(z)

        if z.is_integer:
            if z.is_nonpositive:
                return oo
            elif z.is_positive:
                return log(gamma(z))
        elif z.is_rational:
            p, q = z.as_numer_denom()
            # Half-integral values:
            if p.is_positive and q == 2:
                return log(sqrt(pi) * 2**(1 - p) * gamma(p) / gamma((p + 1) / 2))

        if z is oo:
            return oo
        elif abs(z) is oo:
            return zoo

    def _eval_expand_func(self, **hints):
        from ...concrete import Sum
        z = self.args[0]

        if z.is_Rational:
            p, q = z.as_numer_denom()
            # General rational arguments (u + p/q)
            # Split z as n + p/q with p < q
            n = p // q
            p = p - n*q
            assert p.is_positive and q.is_positive and p < q
            k = Dummy('k')
            if n.is_positive:
                return loggamma(p / q) - n*log(q) + Sum(log((k - 1)*q + p), (k, 1, n))
            elif n.is_negative:
                return loggamma(p / q) - n*log(q) + pi*I*n - Sum(log(k*q - p), (k, 1, -n))

        return self

    def _eval_nseries(self, x, n, logx=None):
        x0 = self.args[0].limit(x, 0)
        if x0 == 0:
            f = self._eval_rewrite_as_intractable(*self.args)
            return f._eval_nseries(x, n, logx)
        return super()._eval_nseries(x, n, logx)

    def _eval_aseries(self, n, args0, x, logx):
        from ...series import Order
        if args0[0] != oo:
            return super()._eval_aseries(n, args0, x, logx)
        z = self.args[0]
        m = min(n, ceiling(Rational(n + 1, 2)))
        r = log(z)*(z - Rational(1, 2)) - z + log(2*pi)/2
        l = [bernoulli(2*k) / (2*k*(2*k - 1)*z**(2*k - 1)) for k in range(1, m)]
        o = None
        if m == 0:
            o = Order(1, x)
        else:
            o = Order(1/z**(2*m - 1), x)
        # It is very inefficient to first add the order and then do the nseries
        return (r + Add(*l))._eval_nseries(x, n, logx) + o

    def _eval_rewrite_as_intractable(self, z):
        return log(gamma(z))

    def _eval_is_extended_real(self):
        if self.args[0].is_nonnegative:
            return True

    def _eval_conjugate(self):
        z = self.args[0]
        if (z.is_extended_real and z.is_nonpositive) is False:
            return self.func(z.conjugate())

    def fdiff(self, argindex=1):
        if argindex == 1:
            return polygamma(0, self.args[0])
        else:
            raise ArgumentIndexError(self, argindex)


def digamma(x):
    r"""
    The digamma function is the first derivative of the loggamma function i.e,

    .. math::
        \psi(x) := \frac{\mathrm{d}}{\mathrm{d} z} \log\Gamma(z)
                = \frac{\Gamma'(z)}{\Gamma(z) }

    In this case, ``digamma(z) = polygamma(0, z)``.

    See Also
    ========

    gamma: Gamma function.
    lowergamma: Lower incomplete gamma function.
    uppergamma: Upper incomplete gamma function.
    polygamma: Polygamma function.
    loggamma: Log Gamma function.
    trigamma: Trigamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Digamma_function
    * https://mathworld.wolfram.com/DigammaFunction.html
    * http://functions.wolfram.com/GammaBetaErf/PolyGamma2/

    """
    return polygamma(0, x)


def trigamma(x):
    r"""
    The trigamma function is the second derivative of the loggamma function i.e,

    .. math::
        \psi^{(1)}(z) := \frac{\mathrm{d}^{2}}{\mathrm{d} z^{2}} \log\Gamma(z).

    In this case, ``trigamma(z) = polygamma(1, z)``.

    See Also
    ========

    gamma: Gamma function.
    lowergamma: Lower incomplete gamma function.
    uppergamma: Upper incomplete gamma function.
    polygamma: Polygamma function.
    loggamma: Log Gamma function.
    digamma: Digamma function.
    diofant.functions.special.beta_functions.beta: Euler Beta function.

    References
    ==========

    * https://en.wikipedia.org/wiki/Trigamma_function
    * https://mathworld.wolfram.com/TrigammaFunction.html
    * http://functions.wolfram.com/GammaBetaErf/PolyGamma2/

    """
    return polygamma(1, x)



from itertools import combinations_with_replacement

from ..core import Add, Dummy, Rational, symbols
from ..polys import (ComputationFailed, Poly, cancel, parallel_poly_from_expr,
                     reduced)
from ..polys.monomials import Monomial
from ..utilities.misc import debug


def ratsimp(expr):
    """
    Put an expression over a common denominator, cancel and reduce.

    Examples
    ========

    >>> ratsimp(1/x + 1/y)
    (x + y)/(x*y)

    """
    f, g = cancel(expr).as_numer_denom()
    try:
        Q, r = reduced(f, [g], field=True, expand=False)
    except ComputationFailed:
        return f/g

    return Add(*Q) + cancel(r/g)


def ratsimpmodprime(expr, G, *gens, **args):
    """
    Simplifies a rational expression ``expr`` modulo the prime ideal
    generated by ``G``.  ``G`` should be a Gröbner basis of the
    ideal.

    >>> eq = (x + y**5 + y)/(x - y)
    >>> ratsimpmodprime(eq, [x*y**5 - x - y], x, y, order='lex')
    (x**2 + x*y + x + y)/(x**2 - x*y)

    If ``polynomial`` is False, the algorithm computes a rational
    simplification which minimizes the sum of the total degrees of
    the numerator and the denominator.

    If ``polynomial`` is True, this function just brings numerator and
    denominator into a canonical form. This is much faster, but has
    potentially worse results.

    References
    ==========

    M. Monagan, R. Pearce, Rational Simplification Modulo a Polynomial
    Ideal,
    http://citeseer.ist.psu.edu/viewdoc/summary?doi=10.1.1.163.6984
    (specifically, the second algorithm)

    """
    from ..matrices import Matrix, zeros
    from ..solvers.solvers import minsolve_linear_system

    quick = args.pop('quick', True)
    polynomial = args.pop('polynomial', False)
    debug('ratsimpmodprime', expr)

    # usual preparation of polynomials:

    num, denom = cancel(expr).as_numer_denom()

    polys, opt = parallel_poly_from_expr([num, denom] + G, *gens, **args)
    domain = opt.domain
    opt.domain = domain.field

    # compute only once
    leading_monomials = [g.LM(opt.order) for g in polys[2:]]
    tested = set()

    def staircase(n):
        """
        Compute all monomials with degree less than ``n`` that are
        not divisible by any element of ``leading_monomials``.

        """
        if n == 0:
            return [1]
        S = []
        for mi in combinations_with_replacement(range(len(opt.gens)), n):
            m = [0]*len(opt.gens)
            for i in mi:
                m[i] += 1
            if all(not lmg.divides(m) for lmg in leading_monomials):
                S.append(m)

        return [Monomial(s).as_expr(*opt.gens) for s in S] + staircase(n - 1)

    def _ratsimpmodprime(a, b, allsol, N=0, D=0):
        r"""
        Computes a rational simplification of ``a/b`` which minimizes
        the sum of the total degrees of the numerator and the denominator.

        The algorithm proceeds by looking at ``a * d - b * c`` modulo
        the ideal generated by ``G`` for some ``c`` and ``d`` with degree
        less than ``a`` and ``b`` respectively.
        The coefficients of ``c`` and ``d`` are indeterminates and thus
        the coefficients of the normalform of ``a * d - b * c`` are
        linear polynomials in these indeterminates.
        If these linear polynomials, considered as system of
        equations, have a nontrivial solution, then `\frac{a}{b}
        \equiv \frac{c}{d}` modulo the ideal generated by ``G``. So,
        by construction, the degree of ``c`` and ``d`` is less than
        the degree of ``a`` and ``b``, so a simpler representation
        has been found.
        After a simpler representation has been found, the algorithm
        tries to reduce the degree of the numerator and denominator
        and returns the result afterwards.

        As an extension, if quick=False, we look at all possible degrees such
        that the total degree is less than *or equal to* the best current
        solution. We retain a list of all solutions of minimal degree, and try
        to find the best one at the end.

        """
        c, d = a, b
        steps = 0

        maxdeg = a.total_degree() + b.total_degree()
        if quick:
            bound = maxdeg - 1
        else:
            bound = maxdeg
        while N + D <= bound:
            if (N, D) in tested:
                break
            tested.add((N, D))

            M1 = staircase(N)
            M2 = staircase(D)
            debug(f'{N} / {D}: {M1}, {M2}')

            Cs = symbols('c:%d' % len(M1), cls=Dummy)
            Ds = symbols('d:%d' % len(M2), cls=Dummy)
            ng = Cs + Ds

            c_hat = Poly(
                sum(Cs[i] * M1[i] for i in range(len(M1))), opt.gens + ng)
            d_hat = Poly(
                sum(Ds[i] * M2[i] for i in range(len(M2))), opt.gens + ng)

            r = reduced(a * d_hat - b * c_hat, G, opt.gens + ng,
                        order=opt.order, polys=True)[1]

            S = Poly(r, gens=opt.gens).coeffs()
            Sv = Cs + Ds
            Sm = zeros(len(S), len(Sv) + 1)
            Sm[:, :-1] = Matrix([[a.diff(b) for b in Sv] for a in S])
            sol = minsolve_linear_system(Sm, *Sv, quick=True)

            if sol and not all(s == 0 for s in sol.values()):
                c = c_hat.subs(sol)
                d = d_hat.subs(sol)

                # The "free" variables occuring before as parameters
                # might still be in the substituted c, d, so set them
                # to the value chosen before:
                c = c.subs(dict(zip(Cs + Ds, [1] * (len(Cs) + len(Ds)))))
                d = d.subs(dict(zip(Cs + Ds, [1] * (len(Cs) + len(Ds)))))

                c = Poly(c, opt.gens)
                d = Poly(d, opt.gens)
                if d == 0:
                    raise ValueError('Ideal not prime?')

                allsol.append((c_hat, d_hat, S, Cs + Ds))
                if N + D != maxdeg:
                    allsol = [allsol[-1]]

                break

            steps += 1
            N += 1
            D += 1

        if steps > 0:
            c, d, allsol = _ratsimpmodprime(c, d, allsol, N, D - steps)
            c, d, allsol = _ratsimpmodprime(c, d, allsol, N - steps, D)

        return c, d, allsol

    # preprocessing. this improves performance a bit when deg(num)
    # and deg(denom) are large:
    num = reduced(num, G, opt.gens, order=opt.order)[1]
    denom = reduced(denom, G, opt.gens, order=opt.order)[1]

    if polynomial:
        return (num/denom).cancel()

    c, d, allsol = _ratsimpmodprime(
        Poly(num, opt.gens, domain=opt.domain), Poly(denom, opt.gens, domain=opt.domain), [])
    if not quick and allsol:
        debug('Looking for best minimal solution. Got: %s' % len(allsol))
        newsol = []
        for c_hat, d_hat, S, ng in allsol:
            Sm = zeros(len(S), len(ng) + 1)
            Sm[:, :-1] = Matrix([[a.diff(b) for b in ng] for a in S])
            sol = minsolve_linear_system(Sm, *ng)
            newsol.append((c_hat.subs(sol), d_hat.subs(sol)))
        c, d = min(newsol, key=lambda x: len(x[0].terms()) + len(x[1].terms()))

    if not domain.is_Field:
        cn, c = c.clear_denoms(convert=True)
        dn, d = d.clear_denoms(convert=True)
    else:
        cn, dn = 1, 1

    cf, c, d = cancel((c, d), opt.gens, order=opt.order)  # canonicalize signs
    r = cf*Rational(cn, dn)

    return (c*r.denominator)/(d*r.numerator)



"""Implementaton of :class:`CharacteristicZero` class."""

from .domain import Domain


__all__ = 'CharacteristicZero',


class CharacteristicZero(Domain):
    """Domain that has infinite number of elements."""

    @property
    def characteristic(self):
        return 0



import math

from ..core import Integer


def egyptian_fraction(r, algorithm='Greedy'):
    """Compute an Egyptian fraction of the rational `r`.

    Returns
    =======

    list
        The list of denominators of an Egyptian
        fraction expansion.

    Parameters
    ==========

    r : Rational
        a positive rational number.
    algorithm : { "Greedy", "Graham Jewett", "Takenouchi", "Golomb" }, optional
        Denotes the algorithm to be used (the default is "Greedy").

    Examples
    ========

    >>> egyptian_fraction(Rational(3, 7))
    [3, 11, 231]
    >>> egyptian_fraction(Rational(3, 7), 'Graham Jewett')
    [7, 8, 9, 56, 57, 72, 3192]
    >>> egyptian_fraction(Rational(3, 7), 'Takenouchi')
    [4, 7, 28]
    >>> egyptian_fraction(Rational(3, 7), 'Golomb')
    [3, 15, 35]
    >>> egyptian_fraction(Rational(11, 5), 'Golomb')
    [1, 2, 3, 4, 9, 234, 1118, 2580]

    See Also
    ========

    diofant.core.numbers.Rational

    Notes
    =====

    Currently the following algorithms are supported:

    1) Greedy Algorithm

       Also called the Fibonacci-Sylvester algorithm.
       At each step, extract the largest unit fraction less
       than the target and replace the target with the remainder.

       It has some distinct properties:

       a) Given `p/q` in lowest terms, generates an expansion of maximum
          length `p`. Even as the numerators get large, the number of
          terms is seldom more than a handful.

       b) Uses minimal memory.

       c) The terms can blow up (standard examples of this are 5/121 and
          31/311).  The denominator is at most squared at each step
          (doubly-exponential growth) and typically exhibits
          singly-exponential growth.

    2) Graham Jewett Algorithm

       The algorithm suggested by the result of Graham and Jewett.
       Note that this has a tendency to blow up: the length of the
       resulting expansion is always ``2**(x/gcd(x, y)) - 1``.

    3) Takenouchi Algorithm

       The algorithm suggested by Takenouchi (1921).
       Differs from the Graham-Jewett algorithm only in the handling
       of duplicates.

    4) Golomb's Algorithm

       A method given by Golumb (1962), using modular arithmetic and
       inverses.  It yields the same results as a method using continued
       fractions proposed by Bleicher (1972).

    If the given rational is greater than or equal to 1, a greedy algorithm
    of summing the harmonic sequence 1/1 + 1/2 + 1/3 + ... is used, taking
    all the unit fractions of this sequence until adding one more would be
    greater than the given number.  This list of denominators is prefixed
    to the result from the requested algorithm used on the remainder.  For
    example, if r is 8/3, using the Greedy algorithm, we get [1, 2, 3, 4,
    5, 6, 7, 14, 420], where the beginning of the sequence, [1, 2, 3, 4, 5,
    6, 7] is part of the harmonic sequence summing to 363/140, leaving a
    remainder of 31/420, which yields [14, 420] by the Greedy algorithm.
    The result of egyptian_fraction(Rational(8, 3), "Golomb") is [1, 2, 3,
    4, 5, 6, 7, 14, 574, 2788, 6460, 11590, 33062, 113820], and so on.

    References
    ==========

    * https://en.wikipedia.org/wiki/Egyptian_fraction
    * https://en.wikipedia.org/wiki/Greedy_algorithm_for_Egyptian_fractions
    * https://www.ics.uci.edu/~eppstein/numth/egypt/conflict.html
    * https://ami.uni-eszterhazy.hu/uploads/papers/finalpdf/AMI_42_from129to134.pdf

    """
    if r <= 0:
        raise ValueError('Value must be positive')

    prefix, rem = egypt_harmonic(r)
    if rem == 0:
        return prefix
    x, y = rem.as_numer_denom()

    if algorithm == 'Greedy':
        return prefix + egypt_greedy(x, y)
    elif algorithm == 'Graham Jewett':
        return prefix + egypt_graham_jewett(x, y)
    elif algorithm == 'Takenouchi':
        return prefix + egypt_takenouchi(x, y)
    elif algorithm == 'Golomb':
        return prefix + egypt_golomb(x, y)
    else:
        raise ValueError('Entered invalid algorithm')


def egypt_greedy(x, y):
    if x == 1:
        return [y]
    else:
        a = (-y) % x
        b = y*(y//x + 1)
        c = math.gcd(a, b)
        if c > 1:
            num, denom = a//c, b//c
        else:
            num, denom = a, b
        return [y//x + 1] + egypt_greedy(num, denom)


def egypt_graham_jewett(x, y):
    l = [y] * x

    # l is now a list of integers whose reciprocals sum to x/y.
    # we shall now proceed to manipulate the elements of l without
    # changing the reciprocated sum until all elements are unique.

    while len(l) != len(set(l)):
        l.sort()  # so the list has duplicates. find a smallest pair
        for i in range(len(l) - 1):  # pragma: no branch
            if l[i] == l[i + 1]:
                break
        # we have now identified a pair of identical
        # elements: l[i] and l[i + 1].
        # now comes the application of the result of graham and jewett:
        l[i + 1] = l[i] + 1
        # and we just iterate that until the list has no duplicates.
        l.append(l[i]*(l[i] + 1))
    return sorted(l)


def egypt_takenouchi(x, y):
    l = [y] * x
    while len(l) != len(set(l)):
        l.sort()
        for i in range(len(l) - 1):  # pragma: no branch
            if l[i] == l[i + 1]:
                break
        k = l[i]
        if k % 2 == 0:
            l[i] = l[i] // 2
            del l[i + 1]
        else:
            l[i], l[i + 1] = (k + 1)//2, k*(k + 1)//2
    return sorted(l)


def egypt_golomb(x, y):
    from ..domains import ZZ
    if x == 1:
        return [y]
    xp = ZZ.invert(int(x), int(y))
    rv = [Integer(xp*y)]
    rv.extend(egypt_golomb((x*xp - 1)//y, xp))
    return sorted(rv)


def egypt_harmonic(r):
    rv = []
    d = Integer(1)
    acc = Integer(0)
    while acc + 1/d <= r:
        acc += 1/d
        rv.append(d)
        d += 1
    return rv, r - acc



import pytest

from diofant import oo, poly, sqrt
from diofant.abc import a, x, y
from diofant.polys.dispersion import dispersion, dispersionset


__all__ = ()


def test_dispersion():
    pytest.raises(ValueError, lambda: dispersionset(poly(x*y, x, y),
                                                    poly(x, x)))
    pytest.raises(ValueError, lambda: dispersionset(poly(x, x),
                                                    poly(y, y)))

    fp = poly(0, x)
    assert sorted(dispersionset(fp)) == [0]

    fp = poly(2, x)
    assert sorted(dispersionset(fp)) == [0]

    fp = poly(x + 1, x)
    assert sorted(dispersionset(fp)) == [0]
    assert dispersion(fp) == 0

    fp = poly((x + 1)*(x + 2), x)
    assert sorted(dispersionset(fp)) == [0, 1]
    assert dispersion(fp) == 1

    fp = poly(x*(x + 3), x)
    assert sorted(dispersionset(fp)) == [0, 3]
    assert dispersion(fp) == 3

    fp = poly((x - 3)*(x + 3), x)
    assert sorted(dispersionset(fp)) == [0, 6]
    assert dispersion(fp) == 6

    fp = poly(x**4 - 3*x**2 + 1, x)
    gp = fp.shift(-3)
    assert sorted(dispersionset(fp, gp)) == [2, 3, 4]
    assert dispersion(fp, gp) == 4
    assert sorted(dispersionset(gp, fp)) == []
    assert dispersion(gp, fp) == -oo

    fp = poly(x**2 + 2*x - 1, x)
    gp = poly(x**2 + 2*x + 3, x)
    assert dispersionset(fp, gp) == set()

    fp = poly(x*(3*x**2+a)*(x-2536)*(x**3+a), x)
    gp = fp.as_expr().subs({x: x - 345}).as_poly(x)
    assert sorted(dispersionset(fp, gp)) == [345, 2881]
    assert sorted(dispersionset(gp, fp)) == [2191]

    gp = poly((x-2)**2*(x-3)**3*(x-5)**3, x)
    assert sorted(dispersionset(gp)) == [0, 1, 2, 3]
    assert sorted(dispersionset(gp, (gp+4)**2)) == [1, 2]

    fp = poly(x*(x+2)*(x-1), x)
    assert sorted(dispersionset(fp)) == [0, 1, 2, 3]

    fp = poly(x**2 + sqrt(5)*x - 1, x, domain='QQ<sqrt(5)>')
    gp = poly(x**2 + (2 + sqrt(5))*x + sqrt(5), x, domain='QQ<sqrt(5)>')
    assert sorted(dispersionset(fp, gp)) == [2]
    assert sorted(dispersionset(gp, fp)) == [1, 4]

    # There are some difficulties if we compute over Z[a]
    # and alpha happenes to lie in Z[a] instead of simply Z.
    # Hence we can not decide if alpha is indeed integral
    # in general.

    fp = poly(4*x**4 + (4*a + 8)*x**3 + (a**2 + 6*a + 4)*x**2 + (a**2 + 2*a)*x, x)
    assert sorted(dispersionset(fp)) == [0, 1]

    # For any specific value of a, the dispersion is 3*a
    # but the algorithm can not find this in general.
    # This is the point where the resultant based Ansatz
    # is superior to the current one.
    fp = poly(a**2*x**3 + (a**3 + a**2 + a + 1)*x, x)
    gp = fp.as_expr().subs({x: x - 3*a}).as_poly(x)
    assert sorted(dispersionset(fp, gp)) == []

    fpa = fp.as_expr().subs({a: 2}).as_poly(x)
    gpa = gp.as_expr().subs({a: 2}).as_poly(x)
    assert sorted(dispersionset(fpa, gpa)) == [6]

    # Work with Expr instead of Poly
    f = (x + 1)*(x + 2)
    assert sorted(dispersionset(f)) == [0, 1]
    assert dispersion(f) == 1

    f = x**4 - 3*x**2 + 1
    g = x**4 - 12*x**3 + 51*x**2 - 90*x + 55
    assert sorted(dispersionset(f, g)) == [2, 3, 4]
    assert dispersion(f, g) == 4

    # Work with Expr and specify a generator
    f = (x + 1)*(x + 2)
    assert sorted(dispersionset(f, None, x)) == [0, 1]
    assert dispersion(f, None, x) == 1

    f = x**4 - 3*x**2 + 1
    g = x**4 - 12*x**3 + 51*x**2 - 90*x + 55
    assert sorted(dispersionset(f, g, x)) == [2, 3, 4]
    assert dispersion(f, g, x) == 4



"""Definitions of common exceptions for :mod:`~diofant.polys` module."""


__all__ = ('BasePolynomialError', 'ExactQuotientFailed',
           'PolynomialDivisionFailed', 'OperationNotSupported',
           'HeuristicGCDFailed', 'HomomorphismFailed',
           'IsomorphismFailed', 'ExtraneousFactors',
           'EvaluationFailed', 'RefinementFailed', 'CoercionFailed',
           'NotInvertible', 'NotReversible', 'NotAlgebraic',
           'DomainError', 'PolynomialError', 'UnificationFailed',
           'GeneratorsError', 'GeneratorsNeeded', 'ComputationFailed',
           'UnivariatePolynomialError', 'MultivariatePolynomialError',
           'PolificationFailed', 'OptionError', 'FlagError')


class BasePolynomialError(Exception):
    """Base class for polynomial related exceptions."""

    def new(self, *args):
        raise NotImplementedError('abstract base class')


class ExactQuotientFailed(BasePolynomialError):
    """Raised when exact quotient is failed."""

    def __init__(self, f, g, dom=None):
        self.f, self.g, self.domain = f, g, dom

    def __str__(self):
        from ..printing import sstr

        if self.domain is None:
            return f'{sstr(self.g)} does not divide {sstr(self.f)}'
        else:
            return f'{sstr(self.g)} does not divide {sstr(self.f)} in {sstr(self.domain)}'

    def new(self, f, g):
        return self.__class__(f, g, self.domain)


class PolynomialDivisionFailed(BasePolynomialError):
    """Raised when polynomial division is failed."""

    def __init__(self, f, g, domain):
        self.f = f
        self.g = g
        self.domain = domain

    def __str__(self):
        if self.domain.is_ExpressionDomain:
            msg = 'You may want to use a different simplification algorithm. Note ' \
                  "that in general it's not possible to guarantee to detect zero "  \
                  'in this domain.'
        elif not self.domain.is_Exact:
            msg = 'Your working precision or tolerance of computations may be set ' \
                  'improperly. Adjust those parameters of the coefficient domain '  \
                  'and try again.'
        else:
            msg = 'Zero detection is guaranteed in this coefficient domain. This '  \
                  'may indicate a bug in Diofant or the domain is user defined and '  \
                  "doesn't implement zero detection properly."

        return "couldn't reduce degree in a polynomial division algorithm when "    \
               f"dividing {self.f} by {self.g}. This can happen when it's not possible to "      \
               'detect zero in the coefficient domain. The domain of computation '  \
               f'is {self.domain}. {msg}'


class OperationNotSupported(BasePolynomialError):
    """Raised when an operation is not supported."""

    def __init__(self, poly, func):
        self.poly = poly
        self.func = func

    def __str__(self):
        return f'`{self.func}` operation not supported by {self.poly.rep.__class__.__name__} representation'


class HeuristicGCDFailed(BasePolynomialError):
    """Raised when a heuristic GCD is failed."""

    pass


class ModularGCDFailed(BasePolynomialError):
    """Raised when a modular GCD is failed."""

    pass


class HomomorphismFailed(BasePolynomialError):
    """Raised when a homomorphism is failed."""

    pass


class IsomorphismFailed(BasePolynomialError):
    """Raised when an isomprphism is failed."""

    pass


class ExtraneousFactors(BasePolynomialError):
    """Raised when there are extraneous factors."""

    pass


class EvaluationFailed(BasePolynomialError):
    """Raised when a polynomial evaluation is failed."""

    pass


class RefinementFailed(BasePolynomialError):
    """Raised when a root refinement is failed."""

    pass


class CoercionFailed(BasePolynomialError):
    """Raised when a coercion is failed."""

    pass


class NotInvertible(BasePolynomialError):
    """Raised when a element is not invertible."""

    pass


class NotReversible(BasePolynomialError):
    """Raised when a element is not reversible."""

    pass


class NotAlgebraic(BasePolynomialError):
    """Raised when a non algebraic element occurred."""

    pass


class DomainError(BasePolynomialError):
    """Generic domain error."""

    pass


class PolynomialError(BasePolynomialError):
    """Generic polynomial error."""

    pass


class UnificationFailed(BasePolynomialError):
    """Raised when domains unification failed."""

    pass


class GeneratorsError(BasePolynomialError):
    """Raised when polynomial generators are unsuitable."""

    pass


class GeneratorsNeeded(GeneratorsError):
    """Raised when more generators needed."""

    pass


class ComputationFailed(BasePolynomialError):
    """Raised when polynomial computation failed."""

    def __init__(self, func, nargs, exc):
        self.func = func
        self.nargs = nargs
        self.exc = exc

    def __str__(self):
        return f'{self.func}({", ".join(map(str, self.exc.exprs[:self.nargs]))}) failed without generators'


class UnivariatePolynomialError(PolynomialError):
    """Generic univariate polynomial error."""

    pass


class MultivariatePolynomialError(PolynomialError):
    """Generic multivariate polynomial error."""

    pass


class PolificationFailed(PolynomialError):
    """Raised if polunomial construction is failed."""

    def __init__(self, opt, origs, exprs, seq=False):
        if not seq:
            self.orig = origs
            self.expr = exprs
            self.origs = [origs]
            self.exprs = [exprs]
        else:
            self.origs = origs
            self.exprs = exprs

        self.opt = opt
        self.seq = seq

    def __str__(self):
        if not self.seq:
            return f"can't construct a polynomial from {self.orig}"
        else:
            return f"can't construct polynomials from {', '.join(map(str, self.origs))}"


class OptionError(BasePolynomialError):
    """Generic option error."""

    pass


class FlagError(OptionError):
    """Generic flag error."""

    pass



"""
This module contain solvers for all kinds of equations,
algebraic or transcendental.
"""

import warnings
from collections import defaultdict
from types import GeneratorType

from ..core import (Add, Dummy, E, Equality, Expr, Float, Function, Ge, I,
                    Integer, Lambda, Mul, Symbol, expand_log, expand_mul,
                    expand_power_exp, nan, nfloat, pi, preorder_traversal,
                    sympify)
from ..core.assumptions import check_assumptions
from ..core.compatibility import (default_sort_key, is_sequence, iterable,
                                  ordered)
from ..core.function import AppliedUndef
from ..core.logic import fuzzy_and
from ..core.relational import Relational
from ..functions import (Abs, Max, Min, Piecewise, acos, arg, asin, atan,
                         atan2, cos, exp, im, log, piecewise_fold, re, sin,
                         sqrt, tan)
from ..functions.elementary.hyperbolic import HyperbolicFunction
from ..functions.elementary.trigonometric import TrigonometricFunction
from ..logic import false, true
from ..matrices import Matrix, zeros
from ..polys import Poly, RootOf, factor, roots
from ..polys.polyerrors import PolynomialError
from ..simplify import (denom, logcombine, nsimplify, posify, powdenest,
                        powsimp, simplify)
from ..simplify.fu import TR1
from ..simplify.sqrtdenest import unrad
from ..utilities import filldedent
from ..utilities.iterables import uniq
from .polysys import solve_linear_system, solve_poly_system, solve_surd_system
from .utils import checksol


__all__ = 'solve', 'solve_linear', 'minsolve_linear_system'


def denoms(eq, symbols=None):
    """Return (recursively) set of all denominators that appear in eq
    that contain any symbol in iterable ``symbols``; if ``symbols`` is
    None (default) then all denominators will be returned.

    Examples
    ========

    >>> denoms(x/y)
    {y}
    >>> denoms(x/(y*z))
    {y, z}
    >>> denoms(3/x + y/z)
    {x, z}
    >>> denoms(x/2 + y/z)
    {2, z}

    """
    pot = preorder_traversal(eq)
    dens = set()
    for p in pot:
        den = denom(p)
        if den == 1:
            continue
        for d in Mul.make_args(den):
            dens.add(d)
    if not symbols:
        return dens
    rv = []
    for d in dens:
        free = d.free_symbols
        if any(s in free for s in symbols):
            rv.append(d)
    return set(rv)


def solve(f, *symbols, **flags):
    r"""Algebraically solves equation or system of equations.

    Parameters
    ==========

    f : Expr, Equality or iterable of above
        All expressions are assumed to be equal to 0.

    \*symbols : tuple
        If none symbols given (empty tuple), free symbols
        of expressions will be used.

    \*\*flags : dict
        A dictionary of following parameters:

        check : bool, optional
            If False, don't do any testing of solutions.  Default is
            True, i.e. the solutions are checked and those that doesn't
            satisfy given assumptions on symbols solved for or make any
            denominator zero - are automatically excluded.
        warn : bool, optional
            Show a warning if :func:`~diofant.solvers.utils.checksol`
            could not conclude.  Default is False.
        simplify : bool, optional
            Enable simplification (default) for all but polynomials of
            order 3 or greater before returning them and (if check is
            not False) use the general simplify function on the solutions
            and the expression obtained when they are substituted into the
            function which should be zero.
        rational : bool or None, optional
            If True, recast Floats as Rational.  If None (default),
            Floats will be recast as rationals but the answer will be
            recast as Floats.  If the flag is False then nothing
            will be done to the Floats.
        cubics, quartics, quintics : bool, optional
            Return explicit solutions (with radicals, which can be quite
            long) when, respectively, cubic, quartic or quintic expressions
            are encountered.  Default is True.  If False,
            :class:`~diofant.polys.rootoftools.RootOf` instances will
            be returned instead.

    Examples
    ========

    Single equation:

    >>> solve(x**2 - y**2)
    [{x: -y}, {x: y}]
    >>> solve(x**2 - 1)
    [{x: -1}, {x: 1}]

    We could restrict solutions by using assumptions:

    >>> p = Symbol('p', positive=True)
    >>> solve(p**2 - 1)
    [{p: 1}]

    Several equations:

    >>> solve((x + 5*y - 2, -3*x + 6*y - 15))
    [{x: -3, y: 1}]
    >>> solve((x + 5*y - 2, -3*x + 6*y - z))
    [{x: -5*z/21 + 4/7, y: z/21 + 2/7}]

    No solution:

    >>> solve([x + 3, x - 3])
    []

    Notes
    =====

    When an object other than a Symbol is given as a symbol, it is
    isolated algebraically and an implicit solution may be obtained.
    This is mostly provided as a convenience to save one from replacing
    the object with a Symbol and solving for that Symbol. It will only
    work if the specified object can be replaced with a Symbol using the
    subs method.

    >>> solve(f(x) - x, f(x))
    [{f(x): x}]
    >>> solve(f(x).diff(x) - f(x) - x, f(x).diff(x))
    [{Derivative(f(x), x): x + f(x)}]

    See Also
    ========

    diofant.solvers.recurr.rsolve : solving recurrence equations
    diofant.solvers.ode.dsolve : solving differential equations
    diofant.solvers.inequalities.reduce_inequalities : solving inequalities

    """
    def _sympified_list(w):
        return list(map(sympify, w if iterable(w) else [w]))
    bare_f = not iterable(f)
    ordered_symbols = (symbols and symbols[0] and
                       (isinstance(symbols[0], (Dummy, Symbol)) or
                        is_sequence(symbols[0], include=GeneratorType)))
    f, symbols = (_sympified_list(w) for w in [f, symbols])

    # preprocess equation(s)
    ###########################################################################
    for i, fi in enumerate(f):
        if isinstance(fi, Equality):
            if 'ImmutableMatrix' in (type(a).__name__ for a in fi.args):
                f[i] = fi.lhs - fi.rhs
            else:
                f[i] = Add(fi.lhs, -fi.rhs, evaluate=False)
        elif isinstance(fi, Relational):
            raise ValueError(f'Only expressions or equalities supported, got {fi}')
        elif isinstance(fi, Poly):
            f[i] = fi.as_expr()

        # rewrite hyperbolics in terms of exp
        f[i] = f[i].replace(lambda w: isinstance(w, HyperbolicFunction),
                            lambda w: w.rewrite(exp))

        # replace min/max:
        f[i] = f[i].replace(lambda w: isinstance(w, (Min, Max)),
                            lambda w: w.rewrite(Piecewise))

        # if we have a Matrix, we need to iterate over its elements again
        if f[i].is_Matrix:
            bare_f = False
            f.extend(list(f[i]))
            f[i] = Integer(0)

        # if we can split it into real and imaginary parts then do so
        freei = f[i].free_symbols
        if freei and all(s.is_extended_real or s.is_imaginary for s in freei):
            fr, fi = f[i].as_real_imag()
            # accept as long as new re, im, arg or atan2 are not introduced
            had = f[i].atoms(re, im, arg, atan2)
            if fr and fi and fr != fi and not any(
                    i.atoms(re, im, arg, atan2) - had for i in (fr, fi)):
                if bare_f:
                    bare_f = False
                f[i: i + 1] = [fr, fi]

    # preprocess symbol(s)
    ###########################################################################
    if not symbols:
        # get symbols from equations
        symbols = set().union(*[fi.free_symbols for fi in f])
        if len(symbols) < len(f):
            for fi in f:
                pot = preorder_traversal(fi)
                for p in pot:
                    if not (p.is_number or p.is_Add or p.is_Mul) or \
                            isinstance(p, AppliedUndef):
                        symbols.add(p)
                        pot.skip()  # don't go any deeper
        symbols = list(symbols)
        # supply dummy symbols so solve(3) behaves like solve(3, x)
        for i in range(len(f) - len(symbols)):
            symbols.append(Dummy())

        ordered_symbols = False
    elif len(symbols) == 1 and iterable(symbols[0]):
        symbols = symbols[0]

    # real/imag handling -----------------------------
    w = Dummy('w')
    piece = Lambda(w, Piecewise((w, Ge(w, 0)), (-w, True)))
    for i, fi in enumerate(f):
        # Abs
        reps = []
        for a in fi.atoms(Abs):
            if not a.has(*symbols):
                continue
            if a.args[0].is_extended_real is None and a.args[0].is_imaginary is not True:
                raise NotImplementedError(f'solving {a} when the argument '
                                          'is not real or imaginary.')
            reps.append((a, piece(a.args[0]) if a.args[0].is_extended_real else
                         piece(a.args[0]*I)))
        fi = fi.subs(reps)

        # arg
        _arg = [a for a in fi.atoms(arg) if a.has(*symbols)]
        fi = fi.xreplace({a: atan(im(a.args[0])/re(a.args[0])) for a in _arg})

        # save changes
        f[i] = fi

    # see if re(s) or im(s) appear
    irf = []
    for s in symbols:
        if s.is_extended_real or s.is_imaginary:
            continue  # neither re(x) nor im(x) will appear
        # if re(s) or im(s) appear, the auxiliary equation must be present
        if any(fi.has(re(s), im(s)) for fi in f):
            irf.append((s, re(s) + I*im(s)))
    if irf:
        for s, rhs in irf:
            for i, fi in enumerate(f):
                f[i] = fi.xreplace({s: rhs})
            f.append(s - rhs)
            symbols.extend([re(s), im(s)])
        if bare_f:
            bare_f = False
    # end of real/imag handling  -----------------------------

    symbols = list(uniq(symbols))
    if not ordered_symbols:
        # we do this to make the results returned canonical in case f
        # contains a system of nonlinear equations; all other cases should
        # be unambiguous
        symbols = sorted(symbols, key=default_sort_key)

    # we can solve for non-symbol entities by replacing them with Dummy symbols
    symbols_new = []
    symbol_swapped = False
    for i, s in enumerate(symbols):
        if s.is_Symbol:
            s_new = s
        else:
            symbol_swapped = True
            s_new = Dummy(f'X{i:d}')
        symbols_new.append(s_new)

    if symbol_swapped:
        swap_sym = list(zip(symbols, symbols_new))
        f = [fi.subs(swap_sym) for fi in f]
        symbols = symbols_new
        swap_sym = {v: k for k, v in swap_sym}
    else:
        swap_sym = {}

    # this is needed in the next two events
    symset = set(symbols)

    # mask off any Object that we aren't going to invert: Derivative,
    # Integral, etc... so that solving for anything that they contain will
    # give an implicit solution
    seen = set()
    non_inverts = set()
    for fi in f:
        pot = preorder_traversal(fi)
        for p in pot:
            if not isinstance(p, Expr) or isinstance(p, Piecewise):
                pot.skip()
            elif (isinstance(p, bool) or not p.args or p in symset or
                  p.is_Add or p.is_Mul or p.is_Pow or p.is_Function or
                  isinstance(p, RootOf)) and p.func not in (re, im):
                pass
            elif p not in seen:
                seen.add(p)
                if p.free_symbols & symset:
                    non_inverts.add(p)
                    pot.skip()
    del seen
    non_inverts = {d: Dummy() for d in non_inverts}
    f = [fi.subs(non_inverts) for fi in f]

    non_inverts = [(v, k.subs(swap_sym)) for k, v in non_inverts.items()]

    # rationalize Floats
    floats = False
    if flags.get('rational', True) is not False:
        for i, fi in enumerate(f):
            if fi.has(Float):
                floats = True
                f[i] = nsimplify(fi, rational=True)

    # piecewise_fold might cancel denominators, so be sure to check them.
    piecewise_dens = set()

    # Any embedded piecewise functions need to be brought out to the
    # top level so that the appropriate strategy gets selected.
    # However, this is necessary only if one of the piecewise
    # functions depends on one of the symbols we are solving for.
    for i, fi in enumerate(f):
        if any(e.has(*symbols) for e in fi.atoms(Piecewise)):
            piecewise_dens |= denoms(fi, symbols)
            f[i] = piecewise_fold(fi)

    if all(_ == 0 for _ in f):
        return [{}]

    #
    # try to get a solution
    ###########################################################################
    if bare_f and len(symbols) == 1:
        solution = [{symbols[0]: s} for s in _solve(f[0], symbols[0], **flags)]
    else:
        solution = _solve_system(f, symbols, **flags)

    #
    # postprocessing
    ###########################################################################
    # Restore masked-off objects
    if non_inverts:
        solution = [{k: v.subs(non_inverts) for k, v in s.items()}
                    for s in solution]

    # Restore original "symbols" if a dictionary is returned.
    # This is not necessary for
    #   - the single univariate equation case
    #     since the symbol will have been removed from the solution;
    #
    # ** unless there were Derivatives with the symbols, but those were handled
    #    above.
    if symbol_swapped:
        symbols = [swap_sym[k] for k in symbols]
        if solution:
            for i, sol in enumerate(solution):
                solution[i] = {swap_sym[k]: v.subs(swap_sym)
                               for k, v in sol.items()}

    # Get assumptions about symbols, to filter solutions.
    # Note that if assumptions about a solution can't be verified, it is still
    # returned.
    check = flags.get('check', True)

    # restore floats
    if floats and solution and flags.get('rational', None) is None:
        solution = nfloat(solution, exponent=False)

    if check and solution:  # assumption checking
        def test_assumptions(sol):
            return fuzzy_and([check_assumptions(sol[sym], **sym._assumptions)
                              for sym in sol])

        solution = [s for s in solution if test_assumptions(s) is not False]

        warn = flags.get('warn', False)
        got_None = [s for s in solution if not test_assumptions(s)]
        if warn and got_None:
            warnings.warn(filldedent("""
                \tWarning: assumptions concerning following solution(s)
                can't be checked:""" + '\n\t' +
                                     ', '.join(str(s) for s in got_None)))

        solution = [s for s in solution if
                    all(not checksol(den, s, **flags) for den in piecewise_dens)]

    #
    # done
    ###########################################################################

    # Make sure that a list of solutions is ordered in a canonical way.
    solution.sort(key=default_sort_key)

    return solution


def _solve(f, symbol, **flags):
    """Return a checked solution for f in terms of one or more of the
    symbols. A list (possibly empty) should be returned.

    If no method is implemented to solve the equation, a NotImplementedError
    will be raised. In the case that conversion of an expression to a Poly
    gives None a ValueError will be raised.

    """
    not_impl_msg = 'No algorithms are implemented to solve equation %s'

    # /!\ capture this flag then set it to False so that no checking in
    # recursive calls will be done; only the final answer is checked
    flags['check'] = checkdens = check = flags.pop('check', True)

    # build up solutions if f is a Mul
    if f.is_Mul:
        result = set()
        for m in f.args:
            soln = _solve(m, symbol, **flags)
            result.update(set(soln))
        result = list(result)
        if check:
            # all solutions have been checked but now we must
            # check that the solutions do not set denominators
            # in any factor to zero
            dens = denoms(f, [symbol])
            result = [s for s in result if
                      all(not checksol(den, {symbol: s}, **flags) for den in
                          dens)]
        # set flags for quick exit at end; solutions for each
        # factor were already checked and simplified
        check = False
        flags['simplify'] = False

    elif f.is_Piecewise:
        result = set()
        for n, (expr, cond) in enumerate(f.args):
            candidates = _solve(piecewise_fold(expr), symbol, **flags)
            for candidate in candidates:
                if candidate in result:
                    continue
                try:
                    v = (cond == true) or cond.subs({symbol: candidate})
                except TypeError:
                    v = False
                if v != false:
                    # Only include solutions that do not match the condition
                    # of any previous pieces.
                    matches_other_piece = False
                    for other_n, (other_expr, other_cond) in enumerate(f.args):  # pragma: no branch
                        if other_n == n:
                            break
                        try:
                            if other_cond.subs({symbol: candidate}) == true:
                                matches_other_piece = True
                                break
                        except TypeError:
                            pass
                    if not matches_other_piece:
                        v = v == true or v.doit()
                        if isinstance(v, Relational):
                            v = v.canonical
                        result.add(Piecewise(
                            (candidate, v),
                            (nan, True)
                        ))
        check = False
        flags['simplify'] = False
    else:
        # first see if it really depends on symbol and whether there
        # is a linear solution
        f_num, sol = solve_linear(f, symbol)
        if symbol not in f_num.free_symbols:
            return []
        elif f_num.is_Symbol:
            # no need to check but simplify if desired
            if flags.get('simplify', True):
                sol = simplify(sol)
            return [sol]

        result = False  # no solution was obtained
        msg = ''  # there is no failure message

        # Poly is generally robust enough to convert anything to
        # a polynomial and tell us the different generators that it
        # contains, so we will inspect the generators identified by
        # polys to figure out what to do.

        # try to identify a single generator that will allow us to solve this
        # as a polynomial, followed (perhaps) by a change of variables if the
        # generator is not a symbol

        poly = Poly(f_num)
        gens = [g for g in poly.gens if g.has(symbol)]

        def _as_base_q(x):
            """Return (b**e, q) for x = b**(p*e/q) where p/q is the leading
            Rational of the exponent of x, e.g. exp(-2*x/3) -> (exp(x), 3)

            """
            b, e = x.as_base_exp()
            if e.is_Rational:
                return b, e.denominator
            if not e.is_Mul:
                return x, 1
            c, ee = e.as_coeff_Mul()
            if c.is_Rational and c != 1:  # c could be a Float
                return b**ee, c.denominator
            return x, 1

        if len(gens) > 1:
            # If there is more than one generator, it could be that the
            # generators have the same base but different powers, e.g.
            #   >>> Poly(exp(x) + 1/exp(x))
            #   Poly(exp(-x) + exp(x), exp(-x), exp(x), domain='ZZ')
            #
            # If unrad was not disabled then there should be no rational
            # exponents appearing as in
            #   >>> Poly(sqrt(x) + sqrt(sqrt(x)))
            #   Poly(sqrt(x) + x**(1/4), sqrt(x), x**(1/4), domain='ZZ')

            bases, qs = list(zip(*[_as_base_q(g) for g in gens]))
            bases = set(bases)

            if len(bases) > 1 or not all(q == 1 for q in qs):
                funcs = {b for b in bases if b.is_Function}

                trig = {_ for _ in funcs if
                        isinstance(_, TrigonometricFunction)}
                other = funcs - trig
                if not other and len(funcs.intersection(trig)) > 1:
                    newf = TR1(f_num).rewrite(tan)
                    if newf != f_num:
                        # don't check the rewritten form --check
                        # solutions in the un-rewritten form below
                        flags['check'] = False
                        result = _solve(newf, symbol, **flags)
                        flags['check'] = check

                # just a simple case - see if replacement of single function
                # clears all symbol-dependent functions, e.g.
                # log(x) - log(log(x) - 1) - 3 can be solved even though it has
                # two generators.

                if result is False and funcs:
                    funcs = list(ordered(funcs))  # put shallowest function first
                    f1 = funcs[0]
                    t = Dummy('t')
                    # perform the substitution
                    ftry = f_num.subs({f1: t})

                    # if no Functions left, we can proceed with usual solve
                    if not ftry.has(symbol):
                        cv_sols = _solve(ftry, t, **flags)
                        cv_inv = _solve(t - f1, symbol, **flags)[0]
                        sols = []
                        for sol in cv_sols:
                            sols.append(cv_inv.subs({t: sol}))
                        result = list(ordered(sols))

                if result is False:
                    msg = f'multiple generators {gens}'

            else:
                # e.g. case where gens are exp(x), exp(-x)
                u = bases.pop()
                t = Dummy('t')
                inv = _solve(u - t, symbol, **flags)
                # this will be resolved by factor in _tsolve but we might
                # as well try a simple expansion here to get things in
                # order so something like the following will work now without
                # having to factor:
                #
                # >>> eq = (exp(I*(-x-2))+exp(I*(x+2)))
                # >>> eq.subs({exp(x): y})  # fails
                # exp(I*(-x - 2)) + exp(I*(x + 2))
                # >>> eq.expand().subs({exp(x): y})  # works
                # y**I*exp(2*I) + y**(-I)*exp(-2*I)

                def _expand(p):
                    b, e = p.as_base_exp()
                    e = expand_mul(e)
                    return expand_power_exp(b**e)
                ftry = f_num.replace(lambda w: w.is_Pow, _expand).subs({u: t})
                assert not ftry.has(symbol)
                soln = _solve(ftry, t, **flags)
                sols = []
                for sol in soln:
                    for i in inv:
                        sols.append(i.subs({t: sol}))
                result = list(ordered(sols))

        else:
            # There is only one generator that we are interested in, but
            # there may have been more than one generator identified by
            # polys (e.g. for symbols other than the one we are interested
            # in) so recast the poly in terms of our generator of interest.

            poly = Poly(f_num, gens[0], extension=False)

            # if we aren't on the tsolve-pass, use roots
            if not flags.pop('tsolve', False):
                deg = poly.degree()
                flags['tsolve'] = True
                solvers = {k: flags.get(k, True) for k in
                           ('cubics', 'quartics', 'quintics')}
                soln = roots(poly, **solvers)
                if sum(soln.values()) < deg:
                    # e.g. roots(32*x**5 + 400*x**4 + 2032*x**3 +
                    #            5000*x**2 + 6250*x + 3189) -> {}
                    # so all_roots is used and RootOf instances are
                    # returned *unless* the system is multivariate
                    # or high-order EX domain.
                    soln = poly.all_roots()
                else:
                    soln = list(soln)

                u = poly.gen
                if u != symbol:
                    try:
                        t = Dummy('t')
                        iv = _solve(u - t, symbol, **flags)
                        soln = list(ordered({i.subs({t: s}) for i in iv for s in soln}))
                    except NotImplementedError:
                        # perhaps _tsolve can handle f_num
                        soln = None
                else:
                    check = False  # only dens need to be checked
                if soln is not None:
                    if len(soln) > 2:
                        # if the flag wasn't set then unset it since high-order
                        # results are quite long. Perhaps one could base this
                        # decision on a certain critical length of the
                        # roots. In addition, wester test M2 has an expression
                        # whose roots can be shown to be real with the
                        # unsimplified form of the solution whereas only one of
                        # the simplified forms appears to be real.
                        flags['simplify'] = flags.get('simplify', False)
                    result = soln

    # fallback if above fails
    # -----------------------
    if result is False:
        u = unrad(f_num, symbol)
        if u:
            eq, cov = u
            if cov:
                isym, ieq = cov
                inv = _solve(ieq, symbol, **flags)[0]
                rv = {inv.subs({isym: xi}) for xi in _solve(eq, isym, **flags)}
            else:
                rv = set(_solve(eq, symbol, **flags))
            result = list(ordered(rv))
            # if the flag wasn't set then unset it since unrad results
            # can be quite long or of very high order
            flags['simplify'] = flags.get('simplify', False)

    # try _tsolve
    if result is False:
        flags.pop('tsolve', None)  # allow tsolve to be used on next pass
        try:
            soln = _tsolve(f_num, symbol, **flags)
            if soln is not None:
                result = soln
        except PolynomialError:
            pass
    # ----------- end of fallback ----------------------------

    if result is False:
        raise NotImplementedError('\n'.join([msg, not_impl_msg % f]))

    if flags.get('simplify', True):
        result = list(map(simplify, result))
        # we just simplified the solution so we now set the flag to
        # False so the simplification doesn't happen again in checksol()
        flags['simplify'] = False

    if checkdens:
        # reject any result that makes any denom. affirmatively 0;
        # if in doubt, keep it
        dens = denoms(f, [symbol])
        result = [s for s in result if
                  all(not checksol(d, {symbol: s}, **flags)
                      for d in dens)]
    if check:
        # keep only results if the check is not False
        result = [r for r in result if
                  checksol(f_num, {symbol: r}, **flags) is not False]
    return result


def _solve_system(exprs, symbols, **flags):
    """Return a checked solution for list of exprs in terms of one or more
    of the symbols. A list of dict's (possibly empty) should be returned.

    """
    if len(symbols) != 1 and len(exprs) == 1:
        f = exprs[0]
        soln = None
        free = f.free_symbols
        ex = free - set(symbols)
        if len(ex) != 1:
            ind, dep = f.as_independent(*symbols)
            ex = ind.free_symbols & dep.free_symbols
        # find first successful solution
        failed = []
        got_s = set()
        result = []
        for s in symbols:
            try:
                soln = _solve(f, s, **flags)
                for sol in soln:
                    if got_s and any(ss in sol.free_symbols for ss in got_s):
                        # sol depends on previously solved symbols: discard it
                        continue
                    got_s.add(s)
                    result.append({s: sol})
            except NotImplementedError:
                continue
        if got_s:
            return result

    polys = []
    surds = []
    dens = set()
    failed = []
    result = [{}]
    solved_syms = []
    algebraic = False
    inversions = False
    checkdens = check = flags.get('check', True)

    for j, g in enumerate(exprs):
        dens.update(denoms(g, symbols))
        i, d = _invert(g, *symbols)
        g = d - i
        if exprs[j] not in (+g, -g):
            inversions = True
        g = g.as_numer_denom()[0]

        poly = g.as_poly(*symbols)

        if poly is not None:
            polys.append(poly)
        elif g.is_algebraic_expr(*symbols):
            surds.append(g)
        else:
            failed.append(g)

    if surds:
        result = solve_surd_system([_.as_expr() for _ in polys] +
                                   surds, *symbols)
        solved_syms = list(set().union(*[set(r) for r in result]))
    elif polys and all(p.is_linear for p in polys):
        n, m = len(polys), len(symbols)
        matrix = zeros(n, m + 1)

        for i, poly in enumerate(polys):
            for monom, coeff in poly.terms():
                try:
                    j = monom.index(1)
                    matrix[i, j] = coeff
                except ValueError:
                    matrix[i, m] = -coeff

        # returns a dictionary {symbols: values} or None
        result = solve_linear_system(matrix, *symbols, **flags)
        solved_syms = list(result) if result else []
        result = [result] if result else [{}]
    elif polys:
        result = solve_poly_system(polys, *symbols)
        solved_syms = list(set().union(*[set(r) for r in result]))

    if failed:
        # For each failed equation, see if we can solve for one of the
        # remaining symbols from that equation. If so, we update the
        # solution set and continue with the next failed equation,
        # repeating until we are done or we get an equation that can't
        # be solved.
        def _ok_syms(e, sort=False):
            rv = (e.free_symbols - solved_syms) & legal
            if sort:
                rv = list(rv)
                rv.sort(key=default_sort_key)
            return rv

        solved_syms = set(solved_syms)  # set of symbols we have solved for
        legal = set(symbols)  # what we are interested in

        # sort so equation with the fewest potential symbols is first
        for eq in ordered(failed, lambda _: len(_ok_syms(_))):
            u = Dummy()  # used in solution checking
            newresult = []
            bad_results = []
            got_s = set()
            hit = False
            for r in result:
                # update eq with everything that is known so far
                eq2 = eq.subs(r)
                # if check is True then we see if it satisfies this
                # equation, otherwise we just accept it
                if check and r:
                    b = checksol(u, {u: eq2}, minimal=True)
                    if b is not None:
                        # this solution is sufficient to know whether
                        # it is valid or not so we either accept or
                        # reject it, then continue
                        if b:
                            newresult.append(r)
                        else:
                            bad_results.append(r)
                        continue
                # search for a symbol amongst those available that
                # can be solved for
                ok_syms = _ok_syms(eq2, sort=True)
                if not ok_syms:
                    newresult.append(r)
                    break  # skip as it's independent of desired symbols
                for s in ok_syms:
                    soln = _solve(eq2, s, **flags)
                    # put each solution in r and append the now-expanded
                    # result in the new result list; use copy since the
                    # solution for s in being added in-place
                    for sol in soln:
                        if got_s and any(ss in sol.free_symbols for ss in got_s):
                            # sol depends on previously solved symbols: discard it
                            continue
                        rnew = r.copy()
                        for k, v in r.items():
                            rnew[k] = v.subs({s: sol})
                        # and add this new solution
                        rnew[s] = sol
                        newresult.append(rnew)
                    hit = True
                    got_s.add(s)
                if not hit:  # pragma: no cover
                    raise NotImplementedError(f'could not solve {eq2}')
            else:
                result = newresult
                assert not any(b in bad_results for b in result)
    else:
        algebraic = True

    default_simplify = bool(failed)  # rely on system-solvers to simplify
    if flags.get('simplify', default_simplify):
        for r in result:
            for k in r:
                r[k] = simplify(r[k])
        flags['simplify'] = False  # don't need to do so in checksol now

    if checkdens:
        result = [r for r in result
                  if not any(checksol(d, r, **flags) for d in dens)]

    if check and (inversions or not algebraic):
        result = [r for r in result
                  if not any(checksol(e, r, **flags) is False for e in exprs)]

    return [r for r in result if r]


def solve_linear(f, x):
    r"""
    Solve equation ``f`` wrt variable ``x``.

    Returns
    =======

    tuple
        ``(x, solution)``, if there is a linear solution, ``(0, 1)`` if
        ``f`` is independent of the symbol ``x``, ``(0, 0)`` if solution set
        any denominator of ``f`` to zero or ``(numerator, denominator)``
        of ``f``, if it's a nonlinear expression wrt ``x``.

    Examples
    ========

    >>> solve_linear(1/x - y**2, x)
    (x, y**(-2))
    >>> solve_linear(x**2/y**2 - 3, x)
    (x**2 - 3*y**2, y**2)
    >>> solve_linear(y, x)
    (0, 1)
    >>> solve_linear(1/(1/x - 2), x)
    (0, 0)

    """
    if not x.is_Symbol:
        raise ValueError(f'{x} is not a Symbol')
    f = f.replace(lambda e: e.is_Derivative, lambda e: e.doit())
    n, d = res = f.as_numer_denom()
    poly = n.as_poly(x, extension=False)
    if poly is not None and poly.is_linear:
        a, b = n.expand().coeff(x, 1), n.expand().coeff(x, 0)
        if a != 0 and d.subs({x: -b/a}) != 0:
            res = (x, -b/a)
    if not n.simplify().has(x):
        res = Integer(0), Integer(1)
    if x == res[0] and any(checksol(_, {x: res[1]}) for _ in denoms(f, [x])):
        res = Integer(0), Integer(0)
    return res


def minsolve_linear_system(system, *symbols, **flags):
    r"""Find a particular solution to a linear system.

    In particular, try to find a solution with the minimal possible number
    of non-zero variables. This is a very computationally hard problem.

    Parameters
    ==========

    system : Matrix
        Nx(M+1) matrix, which means it has to be in augmented form.
    \*symbols : list
        List of M Symbol’s.
    \*\*flags : dict
        A dictionary of following parameters:

        quick : boolean, optional
            If True, a heuristic is used.  Otherwise (default) a naive
            algorithm with exponential complexity is used.

    """
    quick = flags.get('quick', False)
    # Check if there are any non-zero solutions at all
    s0 = solve_linear_system(system, *symbols, **flags)
    if not s0 or all(v == 0 for v in s0.values()):
        return s0
    if quick:
        # We just solve the system and try to heuristically find a nice
        # solution.
        s = solve_linear_system(system, *symbols)

        def update(determined, solution):
            delete = []
            for k, v in solution.items():
                solution[k] = v.subs(determined)
                if not solution[k].free_symbols:
                    delete.append(k)
                    determined[k] = solution[k]
            for k in delete:
                del solution[k]

        determined = {}
        update(determined, s)
        while s:
            # NOTE sort by default_sort_key to get deterministic result
            k = max((k for k in s.values()),
                    key=lambda x: (len(x.free_symbols), default_sort_key(x)))
            x = max(k.free_symbols, key=default_sort_key)
            if len(k.free_symbols) != 1:
                determined[x] = Integer(0)
            else:
                val = solve(k)[0][x]
                if val == 0 and all(v.subs({x: val}) == 0 for v in s.values()):
                    determined[x] = Integer(1)
                else:
                    determined[x] = val
            update(determined, s)
        return determined
    else:
        # We try to select n variables which we want to be non-zero.
        # All others will be assumed zero. We try to solve the modified system.
        # If there is a non-trivial solution, just set the free variables to
        # one. If we do this for increasing n, trying all combinations of
        # variables, we will find an optimal solution.
        # We speed up slightly by starting at one less than the number of
        # variables the quick method manages.
        from itertools import combinations
        from ..utilities.misc import debug
        N = len(symbols)
        bestsol = minsolve_linear_system(system, *symbols, quick=True)
        n0 = len([x for x in bestsol.values() if x != 0])
        for n in range(n0 - 1, 1, -1):
            debug(f'minsolve: {n}')
            thissol = None
            for nonzeros in combinations(list(range(N)), n):
                subm = Matrix([system[:, i].T for i in nonzeros] + [system[:, -1].T]).T
                s = solve_linear_system(subm, *[symbols[i] for i in nonzeros])
                if s and not all(v == 0 for v in s.values()):
                    subs = [(symbols[v], Integer(1)) for v in nonzeros]
                    for k, v in s.items():
                        s[k] = v.subs(subs)
                    for sym in symbols:
                        if sym not in s:
                            if symbols.index(sym) in nonzeros:
                                s[sym] = Integer(1)
                            else:
                                s[sym] = Integer(0)
                    thissol = s
                    break
            if thissol is None:
                break
            bestsol = thissol
        return bestsol


# these are functions that have multiple inverse values per period
multi_inverses = {
    sin: lambda x: (asin(x), pi - asin(x)),
    cos: lambda x: (acos(x), 2*pi - acos(x)),
}


def _tsolve(eq, sym, **flags):
    """
    Helper for _solve that solves a transcendental equation with respect
    to the given symbol. Various equations containing powers and logarithms,
    can be solved.

    There is currently no guarantee that all solutions will be returned or
    that a real solution will be favored over a complex one.

    Either a list of potential solutions will be returned or None will be
    returned (in the case that no method was known to get a solution
    for the equation). All other errors (like the inability to cast an
    expression as a Poly) are unhandled.

    Examples
    ========

    >>> _tsolve(3**(2*x + 5) - 4, x)
    [-5/2 + log(2)/log(3), (-5*log(3)/2 + log(2) + I*pi)/log(3)]

    >>> _tsolve(log(x) + 2*x, x)
    [LambertW(2)/2]

    """
    from .bivariate import bivariate_type, _solve_lambert, _filtered_gens

    if 'tsolve_saw' not in flags:
        flags['tsolve_saw'] = []
    if eq in flags['tsolve_saw']:
        return
    else:
        flags['tsolve_saw'].append(eq)

    rhs, lhs = _invert(eq, sym)

    if lhs == sym:
        return [rhs]
    try:
        if lhs.is_Add:
            # it's time to try factoring; powdenest is used
            # to try get powers in standard form for better factoring
            f = factor(powdenest(lhs - rhs))
            if f.is_Mul:
                return _solve(f, sym, **flags)
            if rhs:
                f = logcombine(lhs, force=flags.get('force', True))
                if f.count(log) != lhs.count(log):
                    if isinstance(f, log):
                        return _solve(f.args[0] - exp(rhs), sym, **flags)
                    else:
                        raise NotImplementedError

        elif lhs.is_Pow:
            if lhs.exp.is_Integer and lhs - rhs != eq:
                return _solve(lhs - rhs, sym, **flags)
            elif sym not in lhs.exp.free_symbols:
                return _solve(lhs.base - rhs**(1/lhs.exp), sym, **flags)
            elif not rhs and sym in lhs.exp.free_symbols:
                # f(x)**g(x) only has solutions where f(x) == 0 and g(x) != 0 at
                # the same place
                sol_base = _solve(lhs.base, sym, **flags)
                return list(ordered(set(sol_base) -
                                    set(_solve(lhs.exp, sym, **flags))))
            elif (rhs != 0 and
                  lhs.base.is_positive and
                  lhs.exp.is_extended_real):
                return _solve(lhs.exp*log(lhs.base) - log(rhs), sym, **flags)
            elif lhs.base == 0 and rhs == 1:
                return _solve(lhs.exp, sym, **flags)

        elif lhs.is_Mul and rhs.is_positive:
            llhs = expand_log(log(lhs))
            if llhs.is_Add:
                return _solve(llhs - log(rhs), sym, **flags)

        elif lhs.is_Function and len(lhs.args) == 1 and lhs.func in multi_inverses:
            # sin(x) = 1/3 -> x - asin(1/3) & x - (pi - asin(1/3))
            soln = []
            for i in multi_inverses[lhs.func](rhs):
                soln.extend(_solve(lhs.args[0] - i, sym, **flags))
            return list(ordered(soln))

        rewrite = lhs.rewrite(exp)
        if rewrite != lhs:
            return _solve(rewrite - rhs, sym, **flags)
    except NotImplementedError:
        pass

    # maybe it is a lambert pattern
    if flags.pop('bivariate', True):
        # lambert forms may need some help being recognized, e.g. changing
        # 2**(3*x) + x**3*log(2)**3 + 3*x**2*log(2)**2 + 3*x*log(2) + 1
        # to 2**(3*x) + (x*log(2) + 1)**3
        g = _filtered_gens(eq.as_poly(), sym)
        up_or_log = set()
        for gi in g:
            if gi.is_Pow and gi.base is E or isinstance(gi, log):
                up_or_log.add(gi)
            elif gi.is_Pow:
                gisimp = powdenest(expand_power_exp(gi))
                if gisimp.is_Pow and sym in gisimp.exp.free_symbols:
                    up_or_log.add(gi)
        eq_down = expand_log(expand_power_exp(eq)).subs(
            dict(zip(up_or_log, [0]*len(up_or_log))))
        eq = expand_power_exp(factor(eq_down, deep=True) + (eq - eq_down))
        rhs, lhs = _invert(eq, sym)
        if lhs.has(sym):
            try:
                poly = lhs.as_poly()
                g = _filtered_gens(poly, sym)
                return _solve_lambert(lhs - rhs, sym, g)
            except NotImplementedError:
                # maybe it's a convoluted function
                if len(g) == 2:
                    try:
                        gpu = bivariate_type(lhs - rhs, *g)
                        if gpu is None:
                            raise NotImplementedError
                        g, p, u = gpu
                        flags['bivariate'] = False
                        inversion = _tsolve(g - u, sym, **flags)
                        if inversion:
                            sol = _solve(p, u, **flags)
                            return list(ordered({i.subs({u: s})
                                                 for i in inversion for s in sol}))
                        else:
                            raise NotImplementedError
                    except NotImplementedError:
                        pass
                else:
                    pass

    if flags.pop('force', True):
        flags['force'] = False
        pos, reps = posify(lhs - rhs)
        for u, s in reps.items():
            if s == sym:
                break
        else:
            u = sym
        if pos.has(u):
            try:
                soln = _solve(pos, u, **flags)
                return list(ordered([s.subs(reps) for s in soln]))
            except NotImplementedError:
                pass


def _invert(eq, *symbols, **kwargs):
    """Return tuple (i, d) where ``i`` is independent of ``symbols`` and ``d``
    contains symbols. ``i`` and ``d`` are obtained after recursively using
    algebraic inversion until an uninvertible ``d`` remains. If there are no
    free symbols then ``d`` will be zero. Some (but not necessarily all)
    solutions to the expression ``i - d`` will be related to the solutions of
    the original expression.

    Examples
    ========

    >>> _invert(x - 3)
    (3, x)
    >>> _invert(3)
    (3, 0)
    >>> _invert(2*cos(x) - 1)
    (1/2, cos(x))
    >>> _invert(sqrt(x) - 3)
    (3, sqrt(x))
    >>> _invert(sqrt(x) + y, x)
    (-y, sqrt(x))
    >>> _invert(sqrt(x) + y, y)
    (-sqrt(x), y)
    >>> _invert(sqrt(x) + y, x, y)
    (0, sqrt(x) + y)

    If there is more than one symbol in a power's base and the exponent
    is not an Integer, then the principal root will be used for the
    inversion:

    >>> _invert(sqrt(x + y) - 2)
    (4, x + y)
    >>> _invert(sqrt(x + y) - 2)
    (4, x + y)

    If the exponent is an integer, setting ``integer_power`` to True
    will force the principal root to be selected:

    >>> _invert(x**2 - 4, integer_power=True)
    (2, x)

    """
    eq = sympify(eq)
    free = eq.free_symbols
    if not symbols:
        symbols = free
    if not free & set(symbols):
        return eq, Integer(0)

    dointpow = bool(kwargs.get('integer_power', False))

    lhs = eq
    rhs = Integer(0)
    while True:
        was = lhs
        while True:
            indep, dep = lhs.as_independent(*symbols)

            # dep + indep == rhs
            if lhs.is_Add:
                # this indicates we have done it all
                if indep == 0:
                    break

                lhs = dep
                rhs -= indep

            # dep * indep == rhs
            else:
                # this indicates we have done it all
                if indep == 1:
                    break

                lhs = dep
                rhs /= indep

        # collect like-terms in symbols
        if lhs.is_Add:
            terms = defaultdict(list)
            for a in lhs.args:
                i, d = a.as_independent(*symbols)
                terms[d].append(i)
            if any(len(v) > 1 for v in terms.values()):
                args = []
                for d, i in terms.items():
                    if len(i) > 1:
                        args.append(Add(*i)*d)
                    else:
                        args.append(i[0]*d)
                lhs = Add(*args)

        # if it's a two-term Add with rhs = 0 and two powers we can get the
        # dependent terms together, e.g. 3*f(x) + 2*g(x) -> f(x)/g(x) = -2/3
        if lhs.is_Add and not rhs and len(lhs.args) == 2 and \
                not lhs.is_polynomial(*symbols):
            a, b = ordered(lhs.args)
            ai, ad = a.as_independent(*symbols)
            bi, bd = b.as_independent(*symbols)
            if any(i.is_Pow for i in (ad, bd)):
                a_base, a_exp = ad.as_base_exp()
                b_base, b_exp = bd.as_base_exp()
                if a_base == b_base:
                    # a = -b
                    lhs = powsimp(powdenest(ad/bd))
                    rhs = -bi/ai
                else:
                    rat = ad/bd
                    _lhs = powsimp(ad/bd)
                    if _lhs != rat:
                        lhs = _lhs
                        rhs = -bi/ai
            if ai*bi == -1:
                if all(
                        isinstance(i, Function) for i in (ad, bd)) and \
                        ad.func == bd.func and len(ad.args) == len(bd.args):
                    if len(ad.args) == 1:
                        lhs = ad.args[0] - bd.args[0]
                    else:
                        # should be able to solve
                        # f(x, y) == f(2, 3) -> x == 2
                        # f(x, x + y) == f(2, 3) -> x == 2 or x == 3 - y
                        raise NotImplementedError('equal function with more than 1 argument')

        elif lhs.is_Mul and any(a.is_Pow for a in lhs.args):
            lhs = powsimp(powdenest(lhs))

        if lhs.is_Function:
            if hasattr(lhs, 'inverse') and len(lhs.args) == 1:
                #                    -1
                # f(x) = g  ->  x = f  (g)
                #
                # /!\ inverse should not be defined if there are multiple values
                # for the function -- these are handled in _tsolve
                #
                rhs = lhs.inverse()(rhs)
                lhs = lhs.args[0]
            elif isinstance(lhs, atan2):
                y, x = lhs.args
                lhs = 2*atan(y/(sqrt(x**2 + y**2) + x))

        if lhs.is_Pow and lhs.base is E:
            rhs = log(rhs)
            lhs = lhs.exp

        if rhs and lhs.is_Pow and lhs.exp.is_Integer and lhs.exp < 0:
            lhs = 1/lhs
            rhs = 1/rhs

        # base**a = b -> base = b**(1/a) if
        #    a is an Integer and dointpow=True (this gives real branch of root)
        #    a is not an Integer and the equation is multivariate and the
        #      base has more than 1 symbol in it
        # The rationale for this is that right now the multi-system solvers
        # doesn't try to resolve generators to see, for example, if the whole
        # system is written in terms of sqrt(x + y) so it will just fail, so we
        # do that step here.
        if lhs.is_Pow and (
            lhs.exp.is_Integer and dointpow or not lhs.exp.is_Integer and
                len(symbols) > 1 and len(lhs.base.free_symbols & set(symbols)) > 1):
            rhs = rhs**(1/lhs.exp)
            lhs = lhs.base

        if lhs == was:
            break
    return rhs, lhs



import collections
import copy

from ..core import Dict, Expr, Integer
from ..core.compatibility import as_int, is_sequence
from ..core.logic import fuzzy_and
from ..functions import sqrt
from ..logic import true
from ..utilities.iterables import uniq
from .matrices import MatrixBase, ShapeError, a2idx


class SparseMatrixBase(MatrixBase):
    """A sparse matrix base class."""

    def __init__(self, *args):
        from . import Matrix

        if len(args) == 1 and isinstance(args[0], SparseMatrixBase):
            self.rows = args[0].rows
            self.cols = args[0].cols
            self._smat = dict(args[0]._smat)
            return

        self._smat = {}

        if len(args) == 3:
            self.rows = as_int(args[0])
            self.cols = as_int(args[1])

            if isinstance(args[2], collections.abc.Callable):
                op = args[2]
                for i in range(self.rows):
                    for j in range(self.cols):
                        value = self._sympify(
                            op(self._sympify(i), self._sympify(j)))
                        if value:
                            self._smat[(i, j)] = value
            elif isinstance(args[2], (dict, Dict)):
                # manual copy, copy.deepcopy() doesn't work
                for key in args[2]:
                    v = args[2][key]
                    if v:
                        self._smat[key] = self._sympify(v)
            elif is_sequence(args[2]):
                if len(args[2]) != self.rows*self.cols:
                    raise ValueError(
                        f'List length ({len(args[2])}) != rows*columns ({self.rows*self.cols})')
                flat_list = args[2]
                for i in range(self.rows):
                    for j in range(self.cols):
                        value = self._sympify(flat_list[i*self.cols + j])
                        if value:
                            self._smat[(i, j)] = value
            else:
                raise ValueError('Third argument must be a callable,'
                                 ' dictionary or sequence.')
        else:
            # handle full matrix forms with _handle_creation_inputs
            r, c, _list = Matrix._handle_creation_inputs(*args)
            self.rows = r
            self.cols = c
            for i in range(self.rows):
                for j in range(self.cols):
                    value = _list[self.cols*i + j]
                    if value:
                        self._smat[(i, j)] = value

    def __getitem__(self, key):

        if isinstance(key, tuple):
            i, j = key
            try:
                i, j = self.key2ij(key)
                return self._smat.get((i, j), Integer(0))
            except (TypeError, IndexError):
                if any(isinstance(_, Expr) and not _.is_number for _ in (i, j)):
                    if ((j < 0) == true) or ((j >= self.shape[1]) == true) or \
                       ((i < 0) == true) or ((i >= self.shape[0]) == true):
                        raise ValueError('index out of boundary')
                    from .expressions.matexpr import MatrixElement
                    return MatrixElement(self, i, j)

                if isinstance(i, slice):
                    i = range(self.rows)[i]
                elif is_sequence(i):
                    pass
                else:
                    if i >= self.rows:
                        raise IndexError('Row index out of bounds')
                    i = [i]
                if isinstance(j, slice):
                    j = range(self.cols)[j]
                elif is_sequence(j):
                    pass
                else:
                    if j >= self.cols:
                        raise IndexError('Col index out of bounds')
                    j = [j]
                return self.extract(i, j)

        # check for single arg, like M[:] or M[3]
        if isinstance(key, slice):
            lo, hi = key.indices(len(self))[:2]
            L = []
            for i in range(lo, hi):
                m, n = divmod(i, self.cols)
                L.append(self._smat.get((m, n), Integer(0)))
            return L

        i, j = divmod(a2idx(key, len(self)), self.cols)
        return self._smat.get((i, j), Integer(0))

    def __setitem__(self, key, value):
        raise NotImplementedError

    def copy(self):
        return self._new(self.rows, self.cols, self._smat)

    @property
    def is_Identity(self):
        if not self.is_square:
            return False
        if not all(self[i, i] == 1 for i in range(self.rows)):
            return False
        return len(self._smat) == self.rows

    def tolist(self):
        """Convert this sparse matrix into a list of nested Python lists.

        Examples
        ========

        >>> a = SparseMatrix(((1, 2), (3, 4)))
        >>> a.tolist()
        [[1, 2], [3, 4]]

        When there are no rows then it will not be possible to tell how
        many columns were in the original matrix:

        >>> SparseMatrix(ones(0, 3)).tolist()
        []

        """
        if not self.rows:
            return []
        I, J = self.shape
        return [[self[i, j] for j in range(J)] for i in range(I)]

    def row_list(self):
        """Returns a row-sorted list of non-zero elements of the matrix.

        Examples
        ========

        >>> a = SparseMatrix(((1, 2), (3, 4)))
        >>> a
        Matrix([
        [1, 2],
        [3, 4]])
        >>> a.RL
        [(0, 0, 1), (0, 1, 2), (1, 0, 3), (1, 1, 4)]

        See Also
        ========

        diofant.matrices.sparse.MutableSparseMatrix.row_op
        col_list

        """
        return [tuple(k + (self[k],)) for k in
                sorted(self._smat, key=lambda k: list(k))]

    RL = property(row_list, None, None, 'Alternate faster representation')

    def col_list(self):
        """Returns a column-sorted list of non-zero elements of the matrix.

        Examples
        ========

        >>> SparseMatrix(((1, 2), (3, 4)))
        Matrix([
        [1, 2],
        [3, 4]])
        >>> _.CL
        [(0, 0, 1), (1, 0, 3), (0, 1, 2), (1, 1, 4)]

        See Also
        ========

        diofant.matrices.sparse.MutableSparseMatrix.col_op
        row_list

        """
        return [tuple(k + (self[k],)) for k in sorted(self._smat, key=lambda k: list(reversed(k)))]

    CL = property(col_list, None, None, 'Alternate faster representation')

    def _eval_trace(self):
        """Calculate the trace of a square matrix.

        Examples
        ========

        >>> eye(3).trace()
        3

        """
        trace = Integer(0)
        for i in range(self.cols):
            trace += self._smat.get((i, i), 0)
        return trace

    def _eval_transpose(self):
        """Returns the transposed SparseMatrix of this SparseMatrix.

        Examples
        ========

        >>> a = SparseMatrix(((1, 2), (3, 4)))
        >>> a
        Matrix([
        [1, 2],
        [3, 4]])
        >>> a.T
        Matrix([
        [1, 3],
        [2, 4]])

        """
        tran = self.zeros(self.cols, self.rows)
        for key, value in self._smat.items():
            key = key[1], key[0]  # reverse
            tran._smat[key] = value
        return tran

    def _eval_conjugate(self):
        """Return the by-element conjugation.

        Examples
        ========

        >>> a = SparseMatrix(((1, 2 + I), (3, 4), (I, -I)))
        >>> a
        Matrix([
        [1, 2 + I],
        [3,     4],
        [I,    -I]])
        >>> a.C
        Matrix([
        [ 1, 2 - I],
        [ 3,     4],
        [-I,     I]])

        See Also
        ========

        transpose: Matrix transposition
        H: Hermite conjugation
        D: Dirac conjugation

        """
        conj = self.copy()
        for key, value in self._smat.items():
            conj._smat[key] = value.conjugate()
        return conj

    def multiply(self, other):
        """Fast multiplication exploiting the sparsity of the matrix.

        Examples
        ========

        >>> A, B = SparseMatrix(ones(4, 3)), SparseMatrix(ones(3, 4))
        >>> A.multiply(B) == 3*ones(4)
        True

        See Also
        ========

        add

        """
        A = self
        B = other
        # sort B's row_list into list of rows
        Blist = [[] for i in range(B.rows)]
        for i, j, v in B.row_list():
            Blist[i].append((j, v))
        Cdict = collections.defaultdict(int)
        for k, j, Akj in A.row_list():
            for n, Bjn in Blist[j]:
                temp = Akj*Bjn
                Cdict[k, n] += temp
        rv = self.zeros(A.rows, B.cols)
        rv._smat = {k: v for k, v in Cdict.items() if v}
        return rv

    def scalar_multiply(self, scalar):
        """Scalar element-wise multiplication."""
        M = self.zeros(*self.shape)
        if scalar:
            for i in self._smat:
                M._smat[i] = scalar*self._smat[i]
        return M

    def __mul__(self, other):
        """Multiply self and other, watching for non-matrix entities.

        When multiplying be a non-sparse matrix, the result is no longer
        sparse.

        Examples
        ========

        >>> I = SparseMatrix(eye(3))
        >>> I*I == I
        True
        >>> Z = zeros(3)
        >>> I*Z
        Matrix([
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0]])
        >>> I*2 == 2*I
        True

        """
        if isinstance(other, SparseMatrixBase):
            return self.multiply(other)
        if isinstance(other, MatrixBase):
            return other._new(self*self._new(other))
        return self.scalar_multiply(other)

    def __rmul__(self, other):
        """Return product the same type as other (if a Matrix).

        When multiplying be a non-sparse matrix, the result is no longer
        sparse.

        Examples
        ========

        >>> A = Matrix(2, 2, range(1, 5))
        >>> S = SparseMatrix(2, 2, range(2, 6))
        >>> A*S == S*A
        False
        >>> (isinstance(A*S, SparseMatrix) is
        ...  isinstance(S*A, SparseMatrix) is False)
        True

        """
        return self.scalar_multiply(other)

    def __add__(self, other):
        """Add other to self, efficiently if possible.

        When adding a non-sparse matrix, the result is no longer
        sparse.

        Examples
        ========

        >>> A = SparseMatrix(eye(3)) + SparseMatrix(eye(3))
        >>> B = SparseMatrix(eye(3)) + eye(3)
        >>> A
        Matrix([
        [2, 0, 0],
        [0, 2, 0],
        [0, 0, 2]])
        >>> A == B
        True
        >>> isinstance(A, SparseMatrix) and isinstance(B, SparseMatrix)
        False

        """
        if isinstance(other, SparseMatrixBase):
            return self.add(other)
        elif isinstance(other, MatrixBase):
            return other._new(other + self)
        else:
            return NotImplemented

    def __neg__(self):
        """Negate all elements of self.

        Examples
        ========

        >>> -SparseMatrix(eye(3))
        Matrix([
        [-1,  0,  0],
        [ 0, -1,  0],
        [ 0,  0, -1]])

        """
        rv = self.copy()
        for k, v in rv._smat.items():
            rv._smat[k] = -v
        return rv

    def add(self, other):
        """Add two sparse matrices with dictionary representation.

        Examples
        ========

        >>> SparseMatrix(eye(3)).add(SparseMatrix(ones(3)))
        Matrix([
        [2, 1, 1],
        [1, 2, 1],
        [1, 1, 2]])
        >>> SparseMatrix(eye(3)).add(-SparseMatrix(eye(3)))
        Matrix([
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0]])

        Only the non-zero elements are stored, so the resulting dictionary
        that is used to represent the sparse matrix is empty:

        >>> _._smat
        {}

        See Also
        ========

        multiply

        """
        if not isinstance(other, SparseMatrixBase):
            raise ValueError(f'only use add with {self.__class__.__name__}, not {other.__class__.__name__}')
        if self.shape != other.shape:
            raise ShapeError()
        M = self.copy()
        for i, v in other._smat.items():
            v = M[i] + v
            if v:
                M._smat[i] = v
            else:
                M._smat.pop(i, None)
        return M

    def extract(self, rowsList, colsList):
        urow = list(uniq(rowsList))
        ucol = list(uniq(colsList))
        smat = {}
        if len(urow)*len(ucol) < len(self._smat):
            # there are fewer elements requested than there are elements in the matrix
            for i, r in enumerate(urow):
                for j, c in enumerate(ucol):
                    smat[i, j] = self._smat.get((r, c), 0)
        else:
            # most of the request will be zeros so check all of self's entries,
            # keeping only the ones that are desired
            for rk, ck in self._smat:
                if rk in urow and ck in ucol:
                    smat[(urow.index(rk), ucol.index(ck))] = self._smat[(rk, ck)]

        rv = self._new(len(urow), len(ucol), smat)
        # rv is nominally correct but there might be rows/cols
        # which require duplication
        if len(rowsList) != len(urow):
            for i, r in enumerate(rowsList):
                i_previous = rowsList.index(r)
                if i_previous != i:
                    rv = rv.row_insert(i, rv[i_previous, :])
        if len(colsList) != len(ucol):
            for i, c in enumerate(colsList):
                i_previous = colsList.index(c)
                if i_previous != i:
                    rv = rv.col_insert(i, rv[:, i_previous])
        return rv
    extract.__doc__ = MatrixBase.extract.__doc__

    @property
    def is_hermitian(self):
        """Checks if the matrix is Hermitian.

        In a Hermitian matrix element i,j is the complex conjugate of
        element j,i.

        Examples
        ========

        >>> a = SparseMatrix([[1, I], [-I, 1]])
        >>> a
        Matrix([
        [ 1, I],
        [-I, 1]])
        >>> a.is_hermitian
        True
        >>> a[0, 0] = 2*I
        >>> a.is_hermitian
        False
        >>> a[0, 0] = x
        >>> a.is_hermitian
        >>> a[0, 1] = a[1, 0]*I
        >>> a.is_hermitian
        False

        """
        def cond():
            d = self._smat
            yield self.is_square
            if len(d) <= self.rows:
                yield fuzzy_and(
                    d[i, i].is_extended_real for i, j in d if i == j)
            else:
                yield fuzzy_and(
                    d[i, i].is_extended_real for i in range(self.rows) if (i, i) in d)
            yield fuzzy_and(
                ((self[i, j] - self[j, i].conjugate()).is_zero
                    if (j, i) in d else False) for (i, j) in d)
        return fuzzy_and(i for i in cond())

    def is_symmetric(self, simplify=True):
        """Return True if self is symmetric.

        Examples
        ========

        >>> M = SparseMatrix(eye(3))
        >>> M.is_symmetric()
        True
        >>> M[0, 2] = 1
        >>> M.is_symmetric()
        False

        """
        if simplify:
            return all((k[1], k[0]) in self._smat and
                       not (self[k] - self[(k[1], k[0])]).simplify()
                       for k in self._smat)
        else:
            return all((k[1], k[0]) in self._smat and
                       self[k] == self[(k[1], k[0])] for k in self._smat)

    def has(self, *patterns):
        """Test whether any subexpression matches any of the patterns.

        Examples
        ========

        >>> A = SparseMatrix(((1, x), (0.2, 3)))
        >>> A.has(x)
        True
        >>> A.has(y)
        False
        >>> A.has(Float)
        True

        """
        return any(self[key].has(*patterns) for key in self._smat)

    def applyfunc(self, f):
        """Apply a function to each element of the matrix.

        Examples
        ========

        >>> m = SparseMatrix(2, 2, lambda i, j: i*2+j)
        >>> m
        Matrix([
        [0, 1],
        [2, 3]])
        >>> m.applyfunc(lambda i: 2*i)
        Matrix([
        [0, 2],
        [4, 6]])

        """
        if not callable(f):
            raise TypeError('`f` must be callable.')

        out = self.copy()
        for k, v in self._smat.items():
            fv = f(v)
            if fv:
                out._smat[k] = fv
            else:
                out._smat.pop(k, None)
        return out

    def reshape(self, rows, cols):
        """Reshape matrix while retaining original size.

        Examples
        ========

        >>> S = SparseMatrix(4, 2, range(8))
        >>> S.reshape(2, 4)
        Matrix([
        [0, 1, 2, 3],
        [4, 5, 6, 7]])

        """
        if len(self) != rows*cols:
            raise ValueError(f'Invalid reshape parameters {rows:d} {cols:d}')
        smat = {}
        for k, v in self._smat.items():
            i, j = k
            n = i*self.cols + j
            ii, jj = divmod(n, cols)
            smat[(ii, jj)] = self._smat[(i, j)]
        return self._new(rows, cols, smat)

    def liupc(self):
        """Liu's algorithm, for pre-determination of the Elimination Tree of
        the given matrix, used in row-based symbolic Cholesky factorization.

        Examples
        ========

        >>> S = SparseMatrix([[1, 0, 3, 2],
        ...                   [0, 0, 1, 0],
        ...                   [4, 0, 0, 5],
        ...                   [0, 6, 7, 0]])
        >>> S.liupc()
        ([[0], [], [0], [1, 2]], [4, 3, 4, 4])

        References
        ==========

        Symbolic Sparse Cholesky Factorization using Elimination Trees,
        Jeroen Van Grondelle (1999)

        """
        # Algorithm 2.4, p 17 of reference

        # get the indices of the elements that are non-zero on or below diag
        R = [[] for r in range(self.rows)]
        for r, c, _ in self.row_list():
            if c <= r:
                R[r].append(c)

        inf = len(R)  # nothing will be this large
        parent = [inf]*self.rows
        virtual = [inf]*self.rows
        for r in range(self.rows):
            for c in R[r][:-1]:
                while virtual[c] < r:
                    t = virtual[c]
                    virtual[c] = r
                    c = t
                if virtual[c] == inf:
                    parent[c] = virtual[c] = r
        return R, parent

    def row_structure_symbolic_cholesky(self):
        """Symbolic cholesky factorization, for pre-determination of the
        non-zero structure of the Cholesky factororization.

        Examples
        ========

        >>> S = SparseMatrix([[1, 0, 3, 2],
        ...                   [0, 0, 1, 0],
        ...                   [4, 0, 0, 5],
        ...                   [0, 6, 7, 0]])
        >>> S.row_structure_symbolic_cholesky()
        [[0], [], [0], [1, 2]]

        References
        ==========

        Symbolic Sparse Cholesky Factorization using Elimination Trees,
        Jeroen Van Grondelle (1999)

        """
        R, parent = self.liupc()
        inf = len(R)  # this acts as infinity
        Lrow = copy.deepcopy(R)
        for k in range(self.rows):
            for j in R[k]:
                while j != inf and j != k:
                    Lrow[k].append(j)
                    j = parent[j]
            Lrow[k] = sorted(set(Lrow[k]))
        return Lrow

    def _cholesky_sparse(self):
        """Algorithm for numeric Cholesky factorization of a sparse matrix."""
        Crowstruc = self.row_structure_symbolic_cholesky()
        C = self.zeros(self.rows)
        for i in range(len(Crowstruc)):
            for j in Crowstruc[i]:
                if i != j:
                    C[i, j] = self[i, j]
                    summ = 0
                    for p1 in Crowstruc[i]:  # pragma: no branch
                        if p1 < j:
                            for p2 in Crowstruc[j]:  # pragma: no branch
                                if p2 < j:
                                    if p1 == p2:
                                        summ += C[i, p1]*C[j, p1]
                                else:
                                    break
                        else:
                            break
                    C[i, j] -= summ
                    C[i, j] /= C[j, j]
                else:
                    C[j, j] = self[j, j]
                    summ = 0
                    for k in Crowstruc[j]:  # pragma: no branch
                        if k < j:
                            summ += C[j, k]**2
                        else:
                            break
                    C[j, j] -= summ
                    C[j, j] = sqrt(C[j, j])

        return C

    def _LDL_sparse(self):
        """Algorithm for numeric LDL factization, exploiting sparse structure.
        """
        Lrowstruc = self.row_structure_symbolic_cholesky()
        L = self.eye(self.rows)
        D = self.zeros(self.rows, self.cols)

        for i in range(len(Lrowstruc)):
            for j in Lrowstruc[i]:
                if i != j:
                    L[i, j] = self[i, j]
                    summ = 0
                    for p1 in Lrowstruc[i]:  # pragma: no branch
                        if p1 < j:
                            for p2 in Lrowstruc[j]:  # pragma: no branch
                                if p2 < j:
                                    if p1 == p2:
                                        summ += L[i, p1]*L[j, p1]*D[p1, p1]
                                else:
                                    break
                        else:
                            break
                    L[i, j] -= summ
                    L[i, j] /= D[j, j]
                else:
                    D[i, i] = self[i, i]
                    summ = 0
                    for k in Lrowstruc[i]:  # pragma: no branch
                        if k < i:
                            summ += L[i, k]**2*D[k, k]
                        else:
                            break
                    D[i, i] -= summ

        return L, D

    def _lower_triangular_solve(self, rhs):
        """Fast algorithm for solving a lower-triangular system,
        exploiting the sparsity of the given matrix.

        """
        rows = [[] for i in range(self.rows)]
        for i, j, v in self.row_list():
            if i > j:
                rows[i].append((j, v))
        X = rhs.copy()
        for i in range(self.rows):
            for j, v in rows[i]:
                X[i, 0] -= v*X[j, 0]
            X[i, 0] /= self[i, i]
        return self._new(X)

    def _upper_triangular_solve(self, rhs):
        """Fast algorithm for solving an upper-triangular system,
        exploiting the sparsity of the given matrix.

        """
        rows = [[] for i in range(self.rows)]
        for i, j, v in self.row_list():
            if i < j:
                rows[i].append((j, v))
        X = rhs.copy()
        for i in range(self.rows - 1, -1, -1):
            rows[i].reverse()
            for j, v in rows[i]:
                X[i, 0] -= v*X[j, 0]
            X[i, 0] /= self[i, i]
        return self._new(X)

    def _diagonal_solve(self, rhs):
        """Diagonal solve."""
        return self._new(self.rows, 1, lambda i, j: rhs[i, 0] / self[i, i])

    def _cholesky_solve(self, rhs):
        # for speed reasons, this is not uncommented, but if you are
        # having difficulties, try uncommenting to make sure that the
        # input matrix is symmetric

        # assert self.is_symmetric()
        L = self._cholesky_sparse()
        Y = L._lower_triangular_solve(rhs)
        rv = L.T._upper_triangular_solve(Y)
        return rv

    def _LDL_solve(self, rhs):
        # for speed reasons, this is not uncommented, but if you are
        # having difficulties, try uncommenting to make sure that the
        # input matrix is symmetric

        # assert self.is_symmetric()
        L, D = self._LDL_sparse()
        Z = L._lower_triangular_solve(rhs)
        Y = D._diagonal_solve(Z)
        return L.T._upper_triangular_solve(Y)

    def cholesky(self):
        """
        Returns the Cholesky decomposition L of a matrix A
        such that L * L.T = A

        A must be a square, symmetric, positive-definite
        and non-singular matrix

        Examples
        ========

        >>> A = SparseMatrix(((25, 15, -5), (15, 18, 0), (-5, 0, 11)))
        >>> A.cholesky()
        Matrix([
        [ 5, 0, 0],
        [ 3, 3, 0],
        [-1, 1, 3]])
        >>> A.cholesky() * A.cholesky().T == A
        True

        """
        from ..core import nan, oo
        if not self.is_symmetric():
            raise ValueError('Cholesky decomposition applies only to '
                             'symmetric matrices.')
        M = self.as_mutable()._cholesky_sparse()
        assert not M.has(nan, oo)
        return self._new(M)

    def LDLdecomposition(self):
        """
        Returns the LDL Decomposition (matrices ``L`` and ``D``) of matrix
        ``A``, such that ``L * D * L.T == A``. ``A`` must be a square,
        symmetric, positive-definite and non-singular.

        This method eliminates the use of square root and ensures that all
        the diagonal entries of L are 1.

        Examples
        ========

        >>> A = SparseMatrix(((25, 15, -5), (15, 18, 0), (-5, 0, 11)))
        >>> L, D = A.LDLdecomposition()
        >>> L
        Matrix([
        [   1,   0, 0],
        [ 3/5,   1, 0],
        [-1/5, 1/3, 1]])
        >>> D
        Matrix([
        [25, 0, 0],
        [ 0, 9, 0],
        [ 0, 0, 9]])
        >>> L * D * L.T == A
        True

        """
        from ..core import nan, oo
        if not self.is_symmetric():
            raise ValueError('LDL decomposition applies only to '
                             'symmetric matrices.')
        L, D = self.as_mutable()._LDL_sparse()
        assert not L.has(nan, oo) and not D.has(nan, oo)
        return self._new(L), self._new(D)

    def solve_least_squares(self, rhs, method='LDL'):
        """Return the least-square fit to the data.

        By default the cholesky_solve routine is used (method='CH'); other
        methods of matrix inversion can be used.

        Examples
        ========

        >>> A = Matrix([1, 2, 3])
        >>> B = Matrix([2, 3, 4])
        >>> S = SparseMatrix(A.row_join(B))
        >>> S
        Matrix([
        [1, 2],
        [2, 3],
        [3, 4]])

        If each line of S represent coefficients of Ax + By
        and x and y are [2, 3] then S*xy is:

        >>> r = S*Matrix([2, 3])
        >>> r
        Matrix([
        [ 8],
        [13],
        [18]])

        But let's add 1 to the middle value and then solve for the
        least-squares value of xy:

        >>> xy = S.solve_least_squares(Matrix([8, 14, 18]))
        >>> xy
        Matrix([
        [ 5/3],
        [10/3]])

        The error is given by S*xy - r:

        >>> S*xy - r
        Matrix([
        [1/3],
        [1/3],
        [1/3]])
        >>> _.norm().evalf(2)
        0.58

        If a different xy is used, the norm will be higher:

        >>> xy += ones(2, 1)/10
        >>> (S*xy - r).norm().evalf(2)
        1.5

        See Also
        ========

        diofant.matrices.matrices.MatrixBase.inv

        """
        t = self.T
        return (t*self).inv(method=method)*t*rhs

    def solve(self, rhs, method='LDL'):
        """Return solution to self*soln = rhs using given inversion method.

        See Also
        ========

        diofant.matrices.matrices.MatrixBase.inv

        """
        if not self.is_square:
            if self.rows < self.cols:
                raise ValueError('Under-determined system.')
            else:
                raise ValueError('For over-determined system, M, having '
                                 'more rows than columns, try M.solve_least_squares(rhs).')
        else:
            return self.inv(method=method)*rhs

    def _eval_inverse(self, **kwargs):
        """Return the matrix inverse using Cholesky or LDL (default)
        decomposition as selected with the ``method`` keyword: 'CH' or 'LDL',
        respectively.

        Examples
        ========

        >>> A = SparseMatrix([[+2, -1, +0],
        ...                   [-1, +2, -1],
        ...                   [+0, +0, +2]])
        >>> A.inv('CH')
        Matrix([
        [2/3, 1/3, 1/6],
        [1/3, 2/3, 1/3],
        [  0,   0, 1/2]])
        >>> A.inv(method='LDL')  # use of 'method=' is optional
        Matrix([
        [2/3, 1/3, 1/6],
        [1/3, 2/3, 1/3],
        [  0,   0, 1/2]])
        >>> A * _
        Matrix([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]])

        """
        sym = self.is_symmetric()
        M = self.as_mutable()
        I = M.eye(M.rows)
        if not sym:
            t = M.T
            r1 = M[0, :]
            M = t*M
            I = t*I
        method = kwargs.get('method', 'LDL')
        if method in 'LDL':
            solve = M._LDL_solve
        elif method == 'CH':
            solve = M._cholesky_solve
        else:
            raise NotImplementedError(f'Method may be "CH" or "LDL", not {method}.')
        rv = M.hstack(*[solve(I[:, i]) for i in range(I.cols)])
        if not sym:
            scale = (r1*rv[:, 0])[0, 0]
            rv /= scale
        return self._new(rv)

    def __eq__(self, other):
        try:
            if self.shape != other.shape:
                return False
            if isinstance(other, SparseMatrixBase):
                return self._smat == other._smat
            else:
                return self._smat == MutableSparseMatrix(other)._smat
        except AttributeError:
            return False

    def as_mutable(self):
        """Returns a mutable version of this matrix.

        Examples
        ========

        >>> X = ImmutableMatrix([[1, 2], [3, 4]])
        >>> Y = X.as_mutable()
        >>> Y[1, 1] = 5  # Can set values in Y
        >>> Y
        Matrix([
        [1, 2],
        [3, 5]])

        """
        return MutableSparseMatrix(self)

    def as_immutable(self):
        """Returns an Immutable version of this Matrix."""
        from .immutable import ImmutableSparseMatrix
        return ImmutableSparseMatrix(self)

    def nnz(self):
        """Returns the number of non-zero elements in Matrix."""
        return len(self._smat)

    @classmethod
    def zeros(cls, r, c=None):
        """Return an r x c matrix of zeros, square if c is omitted."""
        c = r if c is None else c
        r = as_int(r)
        c = as_int(c)
        return cls(r, c, {})

    @classmethod
    def eye(cls, n):
        """Return an n x n identity matrix."""
        n = as_int(n)
        return cls(n, n, {(i, i): Integer(1) for i in range(n)})


class MutableSparseMatrix(SparseMatrixBase, MatrixBase):
    """
    A sparse matrix (a matrix with a large number of zero elements).

    Examples
    ========

    >>> SparseMatrix(2, 2, range(4))
    Matrix([
    [0, 1],
    [2, 3]])
    >>> SparseMatrix(2, 2, {(1, 1): 2})
    Matrix([
    [0, 0],
    [0, 2]])

    See Also
    ========

    diofant.matrices.dense.DenseMatrix

    """

    @classmethod
    def _new(cls, *args, **kwargs):
        return cls(*args)

    def as_mutable(self):
        return self.copy()

    def __setitem__(self, key, value):
        """Assign value to position designated by key.

        Examples
        ========

        >>> M = SparseMatrix(2, 2, {})
        >>> M[1] = 1
        >>> M
        Matrix([
        [0, 1],
        [0, 0]])
        >>> M[1, 1] = 2
        >>> M
        Matrix([
        [0, 1],
        [0, 2]])
        >>> M = SparseMatrix(2, 2, {})
        >>> M[:, 1] = [1, 1]
        >>> M
        Matrix([
        [0, 1],
        [0, 1]])
        >>> M = SparseMatrix(2, 2, {})
        >>> M[1, :] = [[1, 1]]
        >>> M
        Matrix([
        [0, 0],
        [1, 1]])


        To replace row r you assign to position r*m where m
        is the number of columns:

        >>> M = SparseMatrix(4, 4, {})
        >>> m = M.cols
        >>> M[3*m] = ones(1, m)*2
        >>> M
        Matrix([
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [2, 2, 2, 2]])

        And to replace column c you can assign to position c:

        >>> M[2] = ones(m, 1)*4
        >>> M
        Matrix([
        [0, 0, 4, 0],
        [0, 0, 4, 0],
        [0, 0, 4, 0],
        [2, 2, 4, 2]])

        """
        rv = self._setitem(key, value)
        if rv is not None:
            i, j, value = rv
            if value:
                self._smat[(i, j)] = value
            elif (i, j) in self._smat:
                del self._smat[(i, j)]

    __hash__ = None

    def __delitem__(self, key):
        """Delete portion of self defined by key.

        Examples
        ========

        >>> M = SparseMatrix([[0, 0], [0, 1]])
        >>> M
        Matrix([
        [0, 0],
        [0, 1]])
        >>> del M[0, :]
        >>> M
        Matrix([[0, 1]])
        >>> del M[:, 1]
        >>> M
        Matrix([[0]])

        """
        i, j = self.key2ij(key)
        newD = {}
        if isinstance(i, int) and j == slice(None):
            k = a2idx(i, self.rows)
            for (i, j) in self._smat:
                if i == k:
                    pass
                elif i > k:
                    newD[i - 1, j] = self._smat[i, j]
                else:
                    newD[i, j] = self._smat[i, j]
            self._smat = newD
            self.rows -= 1
        elif i == slice(None) and isinstance(j, int):
            k = a2idx(j, self.cols)
            for (i, j) in self._smat:
                if j == k:
                    pass
                elif j > k:
                    newD[i, j - 1] = self._smat[i, j]
                else:
                    newD[i, j] = self._smat[i, j]
            self._smat = newD
            self.cols -= 1
        else:
            raise NotImplementedError

    def row_swap(self, i, j):
        """Swap, in place, columns i and j.

        Examples
        ========

        >>> S = SparseMatrix.eye(3)
        >>> S[2, 1] = 2
        >>> S.row_swap(1, 0)
        >>> S
        Matrix([
        [0, 1, 0],
        [1, 0, 0],
        [0, 2, 1]])

        """
        if i > j:
            i, j = j, i
        rows = self.row_list()
        temp = []
        for ii, jj, v in rows:
            if ii == i:
                self._smat.pop((ii, jj))
                temp.append((jj, v))
            elif ii == j:
                self._smat.pop((ii, jj))
                self._smat[i, jj] = v
            elif ii > j:
                break
        for k, v in temp:
            self._smat[j, k] = v

    def col_swap(self, i, j):
        """Swap, in place, columns i and j.

        Examples
        ========

        >>> S = SparseMatrix.eye(3)
        >>> S[2, 1] = 2
        >>> S.col_swap(1, 0)
        >>> S
        Matrix([
        [0, 1, 0],
        [1, 0, 0],
        [2, 0, 1]])

        """
        if i > j:
            i, j = j, i
        rows = self.col_list()
        temp = []
        for ii, jj, v in rows:
            if jj == i:
                self._smat.pop((ii, jj))
                temp.append((ii, v))
            elif jj == j:
                self._smat.pop((ii, jj))
                self._smat[ii, i] = v
            elif jj > j:
                break
        for k, v in temp:
            self._smat[k, j] = v

    def row_join(self, other):
        """Returns B appended after A (column-wise augmenting)::

            [A B]

        Examples
        ========

        >>> A = SparseMatrix(((1, 0, 1), (0, 1, 0), (1, 1, 0)))
        >>> A
        Matrix([
        [1, 0, 1],
        [0, 1, 0],
        [1, 1, 0]])
        >>> B = SparseMatrix(((1, 0, 0), (0, 1, 0), (0, 0, 1)))
        >>> B
        Matrix([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]])
        >>> C = A.row_join(B)
        >>> C
        Matrix([
        [1, 0, 1, 1, 0, 0],
        [0, 1, 0, 0, 1, 0],
        [1, 1, 0, 0, 0, 1]])
        >>> C == A.row_join(Matrix(B))
        True

        Joining at row ends is the same as appending columns at the end
        of the matrix:

        >>> C == A.col_insert(A.cols, B)
        True

        """
        A, B = self, other
        if not self:
            return type(self)(other)
        if not A.rows == B.rows:
            raise ShapeError()
        A = A.copy()
        if not isinstance(B, SparseMatrixBase):
            k = 0
            b = B._mat
            for i in range(B.rows):
                for j in range(B.cols):
                    v = b[k]
                    if v:
                        A._smat[(i, j + A.cols)] = v
                    k += 1
        else:
            for (i, j), v in B._smat.items():
                A._smat[(i, j + A.cols)] = v
        A.cols += B.cols
        return A

    def col_join(self, other):
        """Returns B augmented beneath A (row-wise joining)::

            [A]
            [B]

        Examples
        ========

        >>> A = SparseMatrix(ones(3))
        >>> A
        Matrix([
        [1, 1, 1],
        [1, 1, 1],
        [1, 1, 1]])
        >>> B = SparseMatrix.eye(3)
        >>> B
        Matrix([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]])
        >>> C = A.col_join(B)
        >>> C
        Matrix([
        [1, 1, 1],
        [1, 1, 1],
        [1, 1, 1],
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]])
        >>> C == A.col_join(Matrix(B))
        True

        Joining along columns is the same as appending rows at the end
        of the matrix:

        >>> C == A.row_insert(A.rows, Matrix(B))
        True

        """
        A, B = self, other
        if not self:
            return type(self)(other)
        if not A.cols == B.cols:
            raise ShapeError()
        A = A.copy()
        if not isinstance(B, SparseMatrixBase):
            k = 0
            b = B._mat
            for i in range(B.rows):
                for j in range(B.cols):
                    v = b[k]
                    if v:
                        A._smat[(i + A.rows, j)] = v
                    k += 1
        else:
            for (i, j), v in B._smat.items():
                A._smat[i + A.rows, j] = v
        A.rows += B.rows
        return A

    def copyin_list(self, key, value):
        from . import Matrix

        if not is_sequence(value):
            raise TypeError('`value` must be of type list or tuple.')
        self.copyin_matrix(key, Matrix(value))

    def copyin_matrix(self, key, value):
        # include this here because it's not part of BaseMatrix
        rlo, rhi, clo, chi = self.key2bounds(key)
        shape = value.shape
        dr, dc = rhi - rlo, chi - clo
        if shape != (dr, dc):
            raise ShapeError(
                "The Matrix `value` doesn't have the same dimensions "
                'as the in sub-Matrix given by `key`.')
        if not isinstance(value, SparseMatrixBase):
            for i in range(value.rows):
                for j in range(value.cols):
                    self[i + rlo, j + clo] = value[i, j]
        else:
            if (rhi - rlo)*(chi - clo) < len(self):
                for i in range(rlo, rhi):
                    for j in range(clo, chi):
                        self._smat.pop((i, j), None)
            else:
                for i, j, v in self.row_list():
                    assert rlo <= i < rhi and clo <= j < chi
                    self._smat.pop((i, j), None)
            for k, v in value._smat.items():
                i, j = k
                self[i + rlo, j + clo] = value[i, j]

    def zip_row_op(self, i, k, f):
        """In-place operation on row ``i`` using two-arg functor whose args are
        interpreted as ``(self[i, j], self[k, j])``.

        Examples
        ========

        >>> M = SparseMatrix.eye(3)*2
        >>> M[0, 1] = -1
        >>> M.zip_row_op(1, 0, lambda v, u: v + 2*u)
        >>> M
        Matrix([
        [2, -1, 0],
        [4,  0, 0],
        [0,  0, 2]])

        See Also
        ========

        row_op
        col_op

        """
        self.row_op(i, lambda v, j: f(v, self[k, j]))

    def row_op(self, i, f):
        """In-place operation on row ``i`` using two-arg functor whose args are
        interpreted as ``(self[i, j], j)``.

        Examples
        ========

        >>> M = SparseMatrix.eye(3)*2
        >>> M[0, 1] = -1
        >>> M.row_op(1, lambda v, j: v + 2*M[0, j])
        >>> M
        Matrix([
        [2, -1, 0],
        [4,  0, 0],
        [0,  0, 2]])

        See Also
        ========

        zip_row_op
        col_op

        """
        for j in range(self.cols):
            v = self._smat.get((i, j), Integer(0))
            fv = f(v, j)
            if fv:
                self._smat[(i, j)] = fv
            elif v:
                self._smat.pop((i, j))

    def col_op(self, j, f):
        """In-place operation on col j using two-arg functor whose args are
        interpreted as (self[i, j], i) for i in range(self.rows).

        Examples
        ========

        >>> M = SparseMatrix.eye(3)*2
        >>> M[1, 0] = -1
        >>> M.col_op(1, lambda v, i: v + 2*M[i, 0])
        >>> M
        Matrix([
        [ 2, 4, 0],
        [-1, 0, 0],
        [ 0, 0, 2]])

        """
        for i in range(self.rows):
            v = self._smat.get((i, j), Integer(0))
            fv = f(v, i)
            if fv:
                self._smat[(i, j)] = fv
            elif v:
                self._smat.pop((i, j))

    def fill(self, value):
        """Fill self with the given value.

        Notes
        =====

        Unless many values are going to be deleted (i.e. set to zero)
        this will create a matrix that is slower than a dense matrix in
        operations.

        Examples
        ========

        >>> M = SparseMatrix.zeros(3)
        >>> M
        Matrix([
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0]])
        >>> M.fill(1)
        >>> M
        Matrix([
        [1, 1, 1],
        [1, 1, 1],
        [1, 1, 1]])

        """
        if not value:
            self._smat = {}
        else:
            v = self._sympify(value)
            self._smat = {(i, j): v
                          for i in range(self.rows) for j in range(self.cols)}


SparseMatrix = MutableSparseMatrix
